#!/usr/bin/env python

import argparse
import sys
from textwrap import dedent

import numpy as np
from sqlalchemy import text

import tangos as db
from tangos import config, core, parallel_tasks
from tangos.core import Creator, Simulation
from tangos.core.tracking import TrackData
from tangos.input_handlers import get_named_handler_class
from tangos.log import logger
from tangos.query import get_halo, get_simulation
from tangos.tools.add_simulation import SimulationAdderUpdater


def _add_simulation_timesteps(options):
    handler = options.handler
    output_class = get_named_handler_class(handler).best_matching_handler(options.sim)
    output_object = output_class(options.sim)
    output_object.quicker = options.quicker
    adder = SimulationAdderUpdater(output_object, renumber=not options.no_renumber)
    adder.min_halo_particles = options.min_particles
    adder.max_num_objects = options.max_objects
    adder.scan_simulation_and_add_all_descendants()
def add_simulation_timesteps(options):
    parallel_tasks.launch(_add_simulation_timesteps,  [options])

def flag_duplicates_deprecated(opts):

    session = db.core.get_default_session()

    print("unmark all properties:", session.execute(text("update haloproperties set deprecated=0")).rowcount)
    print("duplicate properties marked:", session.execute(text("update haloproperties set deprecated=1 where id in (SELECT min(id) FROM haloproperties GROUP BY halo_id, name_id HAVING COUNT(*)>1 ORDER BY halo_id, name_id)")).rowcount)

    print("unmark all links:", session.execute(text("update halolink set deprecated=0")).rowcount)
    print("duplicate links marked:", session.execute(text("update halolink set deprecated=1 where id in (SELECT min(id) FROM halolink GROUP BY halo_from_id, halo_to_id, relation_id HAVING COUNT(*)>1 ORDER BY halo_from_id, halo_to_id, weight)")).rowcount)

    session.commit()

def remove_duplicates(options):

    session = db.core.get_default_session()

    # Note: the MySQL documentation states that “You cannot delete from a table and
    # select from the same table in a subquery.”. You can however circumvent this
    # limitation by creating an implicit temporary table.
    # See https://dev.mysql.com/doc/refman/5.6/en/delete.html
    # and https://stackoverflow.com/a/45498/2601223
    # Another approach would have been to use an inner join but unfortunately
    # SQLite does not support them in deletes, so we have to resort to this approach.
    # With MySQL 5.7.6 onwards, this implicit temporary tables may be optimized away
    # leading to the very same error we are trying to avoid. This can be fixed by
    # setting the optimizer_switch off for the query.
    # See https://dev.mysql.com/doc/relnotes/mysql/5.7/en/news-5-7-6.html#mysqld-5-7-6-optimizer
    dialect = session.connection().engine.dialect.dialect_description.split("+")[0].lower()
    if dialect == 'mysql':
        session.execute(text("SET @__optimizations = @@SESSION.optimizer_switch"))
        session.execute(text("SET @@SESSION.optimizer_switch = 'derived_merge=off'"))

    count = session.execute(text(dedent("""
        DELETE FROM haloproperties
        WHERE id NOT IN (
            SELECT * FROM (
                SELECT MAX(id)
                FROM haloproperties
                GROUP BY halo_id, name_id
            ) as t
        )
    """))).rowcount

    count_links = session.execute(text(dedent("""
        DELETE FROM halolink
        WHERE id NOT IN (
            SELECT * FROM (
                SELECT MAX(id)
                FROM halolink
                GROUP BY halo_from_id, halo_to_id, relation_id
            ) as t
        )
    """))).rowcount

    if dialect == 'mysql':
        session.execute(text("SET @@SESSION.optimizer_switch = @__optimizations"))
    print("Deleted %d rows" % count)
    print("Deleted %d links" % count_links)
    session.commit()



def rem_simulation_timesteps(options):
    basename = options.sims
    from tangos.util.terminalcontroller import term

    sim = core.get_default_session().query(Simulation).filter_by(
        basename=basename).first()

    if sim == None:
        print(term.GREEN + "Simulation does not exist", term.NORMAL)
    else:
        print(term.RED + "Delete simulation", sim, term.NORMAL)
        core.get_default_session().delete(sim)


def add_tracker(halo, size=None):

    import pynbody

    try:
        halo = get_halo(halo)
    except:

        sim = get_simulation(halo)
        print("Adding tracker for isolated run", sim)
        halo = None

    if halo is not None:
        # check we can open the tracker file
        hfile = halo.load()
        hfile.physical_units()
        use_iord = True
        try:
            hfile.dm['iord']
        except:
            use_iord = False

        # get the centre

        cen = halo.get('SSC', None)
        if cen is None:
            cen = pynbody.analysis.halo.shrink_sphere_center(hfile.dm)

        hfile.ancestor.dm['pos'] -= cen

        if size is None:
            size = '500 pc'

        size = pynbody.units.Unit(size)
        try:
            size.in_units("kpc")
            X = hfile.ancestor.dm[pynbody.filt.Sphere(size)]
        except:
            size.in_units("kpc km s^-1")
            X = hfile.ancestor.dm[pynbody.filt.LowPass("j2", size ** 2)]
            size = str(size.in_units("kpc km s^-1")) + "_kks"

        if len(X) < 2:
            print("Alert! Track data is too short")
            import pdb
            pdb.set_trace()
        # setup the tracker
        tx = TrackData(halo.timestep.simulation)
        print("Tracker halo ID is", tx.halo_number)
        print("Length of track data is", len(X))
        tx.select(X, use_iord)
        ts_trigger = halo.timestep
    else:
        f = sim.timesteps[0].load()
        tx = TrackData(sim)
        if tx.halo_number != 1:
            print("Already have a tracker for this simulation -> abort")
            return
        print("Tracker halo ID is", tx.halo_number)
        tx.particles = np.array(
            np.arange(0, len(f.dm) - 1, 1), dtype=int)
        tx.use_iord = False
        ts_trigger = None

    core.get_default_session().add(tx)
    tx.create_objects(first_timestep=ts_trigger)
    if halo is not None:
        targ = halo.timestep.halos.filter_by(
            object_typecode=1, halo_number=tx.halo_number).first()
        size = str(size).replace(" ", "")
        halo["dm_central_" + size] = targ
        print("Copying SSC...", end=' ')
        sys.stdout.flush()
        while halo is not None:
            try:
                targ['SSC'] = halo['SSC']
            except KeyError:
                pass
            halo = halo.next
            targ = targ.next
        print("done")
    core.get_default_session().commit()


def grep_run_id(opts):
    matching_runs = core.get_default_session().query(Creator).filter(
        Creator.command_line.like("%" + opts.query + "%")).all()
    for r in matching_runs:
        print(r.id, end=' ')
    print('\n')

def grep_run_info(opts):
    matching_runs = core.get_default_session().query(Creator).filter(
        Creator.command_line.like("%" + opts.query + "%")).all()
    for r in matching_runs:
        r.print_info()
    return matching_runs

def list_recent_runs(opts):
    n = opts.num
    run = core.get_default_session().query(Creator).order_by(
        Creator.id.desc()).limit(n).all()
    for r in run:
        r.print_info()


def _erase_run_content(run):
    run.halolinks.delete()
    run.halos.delete()
    run.properties.delete()
    run.timesteps.delete()
    for s in run.simulations:
        core.get_default_session().delete(s)
    core.get_default_session().commit()
    core.get_default_session().delete(run)
    core.get_default_session().commit()

def _get_user_confirmation():
    try:
        return input("Enter 'yes' to continue, or anything else to abort >").lower() == "yes"
    except EOFError:
        return False

def rem_run(id, confirm=True):
    run = core.get_default_session().query(Creator).filter_by(id=id).first()

    if run is None:
        raise ValueError(" Run %i does not exist" % id)

    print("You want to delete everything created by the following run:")
    run.print_info()

    if (not confirm) or _get_user_confirmation():
        _erase_run_content(run)
        print("OK")
    else:
        print("aborted")

def grep_remove_runs(opts):
    print('Will remove the following run(s): ')
    matching_runs = grep_run_info(opts)

    # Always ask for confirmation, as this action could be very destructive
    print("You want to delete everything created by the above run(s)?")
    print(""">>> type "yes" to continue""")

    if input(":").lower() == "yes":
        for r, run in enumerate(matching_runs[::-1]):
            print(f'Removing run {run.id}')
            _erase_run_content(run)
        print("Done")
    else:
        print("aborted")


def rollback(options):
    if len(options.ids)>0:
        for run_id in options.ids:
            rem_run(run_id, not options.force)
    else:
        most_recent_id = core.get_default_session().query(Creator).order_by(
          Creator.id.desc()).first().id
        rem_run(most_recent_id, not options.force)

def dump_id(options):
    import pynbody

    h = db.get_halo(options.halo).load()

    if options.size:
        pynbody.analysis.halo.center(h,vel=False)
        print("Size=",options.size)
        h = h.ancestor[pynbody.filt.Sphere(str(options.size)+" kpc")]

    if options.family!="":
        h = getattr(h,options.family)

    np.savetxt(options.filename,h['iord'],"%d")


def list_available_properties(options):
    from .. import properties
    all_properties = sorted(properties.all_properties())

    def format_class_name(cl):
        return "%s.%s"%(cl.__module__, cl.__name__)

    def format_handler_name(cl):
        if cl.requires_particle_data:
            return cl.works_with_handler.__name__.center(15)
        else:
            return "live".center(15)

    longest_class_name = max(len(format_class_name(cl)) for cl in properties.all_property_classes())
    print("{} | {} | {}".format("name".rjust(30), "handler".center(15), "property class"))
    print("-"*30+"-+-"+"-"*15+"-+-"+"-"*longest_class_name)
    for p in all_properties:
        classes = properties.all_providing_classes(p)
        print("%s | %.15s | %s"%(p.rjust(30), format_handler_name(classes[0]), format_class_name(classes[0])))
        for additional_class in classes[1:]:
            print(" "*30+" | %.15s | %s"%(format_handler_name(additional_class),
                                          format_class_name(additional_class)))

def diff(options):
    from ..testing import db_diff
    differ = db_diff.TangosDbDiff(options.uri1, options.uri2, ignore_keys=options.ignore_value_of)
    if options.property_tolerance is not None:
        for k, rtol, atol in options.property_tolerance:
            differ.set_tolerance(k, float(rtol), float(atol))

    if options.simulation:
        differ.compare_simulation(options.simulation)
    elif options.timestep:
        differ.compare_timestep(options.timestep)
    elif options.object:
        differ.compare_object(options.object)
    else:
        differ.compare()
    result = differ.failed

    if result:
        logger.info("Differences found. Exiting with status 1.")
        sys.exit(1)
    else:
        logger.info("No differences found.")


def main():
    print("""
    The 'tangos_manager' command line is deprecated in favour of just 'tangos'.
    'tangos_manager' may be removed in future versions.
    """)

    parser, _ = get_argument_parser_and_subparsers()

    args = parser.parse_args()
    core.process_options(args)
    args.func(args)


def get_argument_parser_and_subparsers():
    parser = argparse.ArgumentParser()
    core.supplement_argparser(parser)
    subparse = parser.add_subparsers(required=True)

    subparse_add = subparse.add_parser("add",
                                       help="Add new simulations to the database, or update existing simulations")
    subparse_add.add_argument("sim", action="store",
                              help="The path to the simulation folders relative to the database folder")
    subparse_add.add_argument("--handler", action="store",
                              help="The handler to use from the simulation_outputs subpackage",
                              default=config.default_fileset_handler_class)
    subparse_add.add_argument("--min-particles", action="store", type=int, default=config.min_halo_particles,
                              help="The minimum number of particles a halo must have before it is imported (default %d)" % config.min_halo_particles)
    subparse_add.add_argument("--max-objects", action="store", type=int, default=config.max_num_objects,
                              help="The maximum number of objects of a particular type to store (no limit if not specified)")
    subparse_add.add_argument("--quicker", action="store_true",
                              help="Cut corners/make guesses to import quickly and with minimum memory usage. Only use if you understand the consequences!")
    subparse_add.add_argument("--no-renumber", action="store_true",
                              help="By default tangos renumbers halos to start from 1, in decreasing order of dark matter particles. Set this flag to keep the original halo finder numbers.")
    subparse_add.set_defaults(func=add_simulation_timesteps)

    subparse_recentruns = subparse.add_parser("recent-runs",
                                              help="List information about the most recent database updates")
    subparse_recentruns.set_defaults(func=list_recent_runs)
    subparse_recentruns.add_argument("num", type=int,
                                     help="The number of runs to display, starting with the most recent")

    subparse_greprun = subparse.add_parser("grep-run-ids",
                                          help="List IDs of runs matching command line input")
    subparse_greprun.set_defaults(func=grep_run_id)
    subparse_greprun.add_argument("query", type=str,
                                     help="The sub-string to search for in the command line")

    subparse_grepruninfo = subparse.add_parser("grep-runs",
                                          help="List details of runs matching command line input")
    subparse_grepruninfo.set_defaults(func=grep_run_info)
    subparse_grepruninfo.add_argument("query", type=str,
                                     help="The sub-string to search for in the command line")

    subparse_grepremove = subparse.add_parser("grep-remove",
                                           help="Remove runs matching command line input")
    subparse_grepremove.set_defaults(func=grep_remove_runs)
    subparse_grepremove.add_argument("query", type=str,
                                      help="The sub-string to search for in the command line")


    # The following subcommands currently do not work and is disabled:
    """
    subparse_remruns = subparse.add_parser("rm", help="Remove a simulation from the database")
    subparse_remruns.add_argument("sims", help="The path to the simulation folder relative to the database folder")
    subparse_remruns.set_defaults(func=rem_simulation_timesteps)
     """

    subparse_deprecate = subparse.add_parser("flag-duplicates",
                                             help="Flag old copies of properties and duplicate links (if they are present)")
    subparse_deprecate.set_defaults(func=flag_duplicates_deprecated)
    subparse_deprecate = subparse.add_parser("remove-duplicates",
                                             help="Remove old copies of properties and duplicate links (if they are present)")
    subparse_deprecate.set_defaults(func=remove_duplicates)

    subparse_rollback = subparse.add_parser("rollback", help="Remove database updates")
    subparse_rollback.add_argument("ids", nargs="*", type=int, help="IDs of the database updates to remove. If none specified, removes the most recent run.")
    subparse_rollback.add_argument("--force", "-f", action="store_true", help="If this flag is present, no confirmation prompts will be issued")
    subparse_rollback.set_defaults(func=rollback)

    subparse_dump_id = subparse.add_parser("dump-iord", help="Dump the iords corresponding to a specified halo")
    subparse_dump_id.add_argument("halo", type=str, help="The identity of the halo to dump")
    subparse_dump_id.add_argument("filename", type=str, help="A filename for the output text file")
    subparse_dump_id.add_argument("size", type=str, nargs="?",
                                  help="Size, in kpc, of sphere to extract (or omit to get just the halo particles)")
    subparse_dump_id.add_argument("family", type=str, help="The family of particles to extract", default="")
    subparse_dump_id.set_defaults(func=dump_id)


    subparse_diff = subparse.add_parser("diff", help="Analyse the difference between two databases")
    subparse_diff.add_argument("uri1", type=str, help="The first database URI or filename")
    subparse_diff.add_argument("uri2", type=str, help="The second database URI or filename")
    subparse_diff.add_argument("--simulation", type=str, help="Only compare the specified simulation", default=None)
    subparse_diff.add_argument("--timestep", type=str, help="Only compare the specified timestep", default=None)
    subparse_diff.add_argument("--object", type=str, help="Only compare the specified object", default=None)
    subparse_diff.add_argument("--ignore-value-of", nargs="*", type=str, help="Ignore the value of the specified properties", default=[])
    subparse_diff.add_argument("--property-tolerance", nargs=3, type=str, help="Set the relative and absolute tolerances "
                                                                                 "for a given property, as --property-tolerance "
                                                                                 "<property_name> <rtol> <atol> ", action="append")
    subparse_diff.set_defaults(func=diff)

    subparse_list_available_properties = subparse.add_parser("list-possible-properties",
                                                             help="List all the object properties that can be calculated by the currently available modules")
    subparse_list_available_properties.set_defaults(func=list_available_properties)
    return parser, subparse



from .message import Message

FILENAME = "parallel_tasks_test_log.txt"

def initialise_log():
    with open(FILENAME, "w") as f:
        f.write("")

def get_log():
    with open(FILENAME) as f:
        return f.readlines()

def log(message):
    ServerLogMessage(message).send(0)

class ServerLogMessage(Message):
    def process(self):
        with open(FILENAME, "a") as f:
            f.write(f"[{self.source:d}] {self.contents:s}\r\n")



import numpy as np

import tangos as db

from .. import LivePropertyCalculation, PropertyCalculation


class FindCenter(PropertyCalculation):
    """Returns center array in halo finder units"""
    names = "center"

    def requires_property(self):
        return ["X", "Y", "Z"]

    def calculate(self, particle_data, existing_properties):
        return np.array([existing_properties["X"],existing_properties["Y"],existing_properties["Z"]])

class FindPhyiscalCenter(PropertyCalculation):
    """Returns center arrays in physical units (Mpc)"""
    names = "center_Mpc"

    def requires_property(self):
        return ["X_Mpc", "Y_Mpc", "Z_Mpc"]

    def calculate(self, particle_data, existing_properties):
        return np.array([existing_properties["X_Mpc"],existing_properties["Y_Mpc"],existing_properties["Z_Mpc"]])

class FindHosts(LivePropertyCalculation):
    """Returns all more massive halos that given halo is within
    the virial radius of"""
    names = "hosts"

    def requires_property(self):
        return ["Mvir", "center_Mpc", "Rvir_kpc"]

    def live_calculate(self, halo_entry):
        centers, radii, dbid, masses = halo_entry.timestep.calculate_all("center_Mpc","Rvir_kpc","dbid()","Mvir")
        offsets = np.linalg.norm(halo_entry['center_Mpc'] - centers[masses>halo_entry['Mvir']], axis=1)
        host_mask = offsets<(radii[masses>halo_entry['Mvir']]/1000.)
        potids = dbid[masses>halo_entry['Mvir']]
        return np.array([db.get_halo(x) for x in potids[host_mask]])

class FindSats(LivePropertyCalculation):
    """Returns all less massive halos that lie within the virial
    radius of a given halo"""
    names = "satellites"

    def requires_property(self):
        return ["Mvir", "center_Mpc", "Rvir_kpc"]

    def live_calculate(self, halo_entry):
        centers, radii, dbid, masses = halo_entry.timestep.calculate_all("center_Mpc","Rvir_kpc","dbid()","Mvir")
        offsets = np.linalg.norm(halo_entry['center_Mpc'] - centers[masses<halo_entry['Mvir']], axis=1)
        host_mask = offsets<(halo_entry['Rvir_kpc']/1000.)
        potids = dbid[masses<halo_entry['Mvir']]
        return np.array([db.get_halo(x) for x in potids[host_mask]])



class MessageMetaClass(type):
    _message_classes = {}
    _next_tag = 100

    def __new__(meta, name, bases, dct):
        return super().__new__(meta, name, bases, dct)

    def __init__(cls, name, bases, dct):
        super().__init__(name, bases, dct)
        MessageMetaClass.register_class(cls, MessageMetaClass._next_tag)
        MessageMetaClass._next_tag+=1

    @staticmethod
    def tag_to_class(tag):
        if tag not in MessageMetaClass._message_classes:
            raise RuntimeError("Unknown message receieved (tag %d)" % tag)

        return MessageMetaClass._message_classes[tag]

    @staticmethod
    def class_is_known(cls):
        return cls in MessageMetaClass._message_classes.values()

    @staticmethod
    def register_class(cls, tag):
        if MessageMetaClass.class_is_known(cls):
            raise AttributeError("Attempting to register duplicate message class %r"%cls)
        MessageMetaClass._message_classes[tag] = cls
        cls._tag = tag


class Message(metaclass=MessageMetaClass):
    _handler = None

    def __init__(self, contents=None):
        self.contents = contents
        self.source = None

    @classmethod
    def deserialize(cls, source, message):
        obj = cls(message)
        obj.source = source
        return obj

    def serialize(self):
        return self.contents

    @staticmethod
    def interpret_and_deserialize(tag, source, message):
        obj = MessageMetaClass.tag_to_class(tag).deserialize(source, message)
        return obj

    @staticmethod
    def process_incoming_message(tag, source, message):
        obj = Message.interpret_and_deserialize(tag, source, message)
        obj.process()

    def send(self, destination):
        from . import backend
        backend.send(self.serialize(), destination=destination, tag=self._tag)

    @classmethod
    def receive(cls, source=None):
        from . import backend

        msg, source, tag = backend.receive_any(source=None)
        obj = Message.interpret_and_deserialize(tag, source, msg)

        if not isinstance(obj, cls):
            if hasattr(obj, "_is_exception"):
                raise obj.contents
            else:
                raise RuntimeError("Unexpected message of type %r received"%type(obj))
        return obj

    def process(self):
        raise NotImplementedError("No process implemented for this message")


class ExceptionMessage(Message):
    _is_exception = True

    def process(self):
        raise self.contents



"""Tangos - the agile numerical galaxy organization system

This package provides ways to create and interpret databases of numerical galaxy formation simulations for scientific
analysis.

For information on getting started, see README.md.

"""

import sqlalchemy
import sqlalchemy.orm.session
from sqlalchemy import (Boolean, Column, DateTime, Float, ForeignKey, Index,
                        Integer, String, and_, create_engine, or_, orm)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import (backref, clear_mappers, deferred, relationship,
                            sessionmaker)
from sqlalchemy.orm.session import Session

from . import core, log, properties
from .core import *
from .query import *

__version__ = '1.9.1'



"""input_handlers sub-package

This sub-package handles returning views of the original simulation data on disk for further processing.
At the moment, the views are always pynbody views, but adaptors for other frameworks could be implemented by
overriding the functionality.

For an introduction, see https://pynbody.github.io/tangos/input_handlers.html
"""

import importlib
import os
import os.path
import warnings
import weakref

from .. import config
from ..log import logger


class DummyTimeStep:
    def __init__(self, filename):
        self.filename = filename


    def __repr__(self):
        return self.filename

    pass


_loaded_timesteps = weakref.WeakValueDictionary()


class HandlerBase:
    """This class handles the output from a simulation as it resides on disk.

    Subclasses provide implementations for different formats and situations.
    """

    halo_stat_file_class_name = "HaloStatFile"

    def __init__(self, basename):
        self.basename = self.strip_slashes(basename)
        self.quicker = False # a flag to indicate that corners may be cut in the interest of efficiency

    @classmethod
    def best_matching_handler(cls, basename):
        """Find the best subclass to read in the specified folder of simulation timesteps"""
        return cls

    def enumerate_timestep_extensions(self, parallel=False):
        """Yield the extension of each timestep available on disk.

        If parallel is True, then each timestep is returned to one and only one rank."""
        raise NotImplementedError

    def get_properties(self):
        """Returns a dictionary of properties of the simulation"""
        return {}

    def get_timestep_properties(self, ts_extension):
        """Returns a dictionary of properties of the timestep"""
        return {}

    def enumerate_objects(self, ts_extension, object_typetag='halo', min_halo_particles=None):
        """Yield halo_number, NDM, NStar, Ngas for halos in the specified timestep"""
        yield from self._enumerate_objects_from_statfile(ts_extension, object_typetag)

    def _enumerate_objects_from_statfile(self, ts_extension, object_typetag):
        """Implementation of enumerate_objects when the information is provided by a file readable
        by the halo_stat_files module.

        Call from subclasses when this behaviour is desired"""
        statfile = self.get_stat_file(ts_extension, object_typetag)
        logger.info("Reading halos for timestep %r using a stat file", ts_extension)
        yield from statfile.iter_rows("n_dm", "n_star", "n_gas")

    def _can_enumerate_objects_from_statfile(self, ts_extension, object_typetag):
        """Returns True if the objects can be enumerated from a stat file"""
        try:
            self.get_stat_file(ts_extension, object_typetag)
            return True
        except OSError:
            return False

    def get_stat_file(self, ts_extension, object_typetag):
        from . import halo_stat_files
        if object_typetag != 'halo':
            raise OSError("No stat file known for object type %s"%object_typetag)
        #ts = DummyTimeStep()
        #ts.redshift = self.get_timestep_properties(ts_extension)['redshift']
        from . import caterpillar
        statfile = getattr(halo_stat_files, self.halo_stat_file_class_name)(self._extension_to_filename(ts_extension))
        return statfile


    def available_object_property_names_for_timestep(self, ts_extension, object_typetag):
        """Returns a list of all pre-computed properties available for this timestep.

        These pre-computed properties can then be evaluated through iterate_object_properties_for_timestep"""
        return self.get_stat_file(ts_extension, object_typetag).all_columns()

    def iterate_object_properties_for_timestep(self, ts_extension, object_typetag, property_names):
        """Iterate through all objects of specified type, providing named pre-computed data.

        This is normally data that was calculated and stored by the halo finder such as masses or particle counts.
        Each object yields an array starting with the finder offset, followed by the finder_id,
        followed by values for the requested properties. See also method iter_row_raw in halo_stat_file

        :arg ts_extension - the timestep path (relative to the simulation basename)
        :arg object_typetag - the type of halo catalogue (e.g. 'halo' or 'group')
        :arg property_names - a list of property names to retrieve (depends on catalogue file format)
        """
        statfile = self.get_stat_file(ts_extension, object_typetag)
        yield from statfile.iter_rows(*property_names)


    def load_timestep(self, ts_extension, mode=None):
        """Returns an object that connects to the data for a timestep on disk -- possibly a version cached in
        memory"""
        ts_hash = hash((self._extension_to_filename(ts_extension),mode,type(self)))
        stored_timestep = _loaded_timesteps.get(ts_hash, None)
        if stored_timestep is not None:
            return stored_timestep
        else:
            data = self.load_timestep_without_caching(ts_extension, mode=mode)
            _loaded_timesteps[ts_hash] = data
            return data

    def load_region(self, ts_extension, region_specification, mode=None):
        """Returns an object that connects to the data for a timestep on disk, filtered using the
        specified region specification. Acceptable region specifications are output handler dependent.

        The returned object may be a sub-view of a cached in-memory timestep."""
        raise NotImplementedError

    def load_timestep_without_caching(self, ts_extension, mode=None):
        """Creates and returns an object that connects to the data for a timestep on disk"""
        raise NotImplementedError

    def load_object(self, ts_extension, finder_id, catalog_pos, object_typetag='halo', mode=None):
        """Creates and returns an object that connects to the data for a halo on disk.

        :arg ts_extension - the timestep path (relative to the simulation basename)

        :arg finder_id - the halo id in the raw halo finder output

        :arg catalog_pos - the position of the raw halo finder output

        :arg object_typetag - the type of halo catalogue (normally 'halo' or, for Subfind, might be 'group' to get
                                                     the top-level groups)

        :arg mode - sets a method for loading the halo, which is dependent on being implemented by the underlying
                    file manager. Current options (for the default pynbody handler) are

                     * None: load entire snapshot from disk, then return halo particles
                     * 'partial': load only the halo particles from disk
                     * 'server': use a remote server model to retrieve the particles
        """
        raise NotImplementedError

    def load_tracked_region(self, ts_extension, track_data, mode=None):
        """Creates and returns an object that connects to the on-disk data for the specified tracked region.

        :arg mode - sets a method for loading the tracked region; see load_object mode for more information"""
        raise NotImplementedError


    @classmethod
    def handler_class_name(cls):
        module = cls.__module__
        if module.startswith(HandlerBase.__module__):
            submodule = module[len(HandlerBase.__module__)+1:]
            return submodule+"."+cls.__name__
        else:
            return module+"."+cls.__name__

    @staticmethod
    def strip_slashes(name):
        """Strip trailing and leading slashes from relative path"""
        if len(name) == 0: return name
        while name[0] == "/":
            name = name[1:]
            if len(name) == 0: return name
        while name[-1] == "/":
            name = name[:-1]
            if len(name) == 0: return name
        return name

    def _extension_to_filename(self, ts_extension):
        """Given the timestep extension, form the path to the actual file"""
        # Explicit str cast - pynbody seems inconsistent about how it responds to unicode filenames
        return str(os.path.join(config.base, self.basename, ts_extension))


def _map_deprecated_handler_name(handler):
    if 'OutputSetHandler' in handler:
        new_handler = handler.replace('OutputSetHandler','InputHandler')
        warnings.warn("The database has stored the handler name as %r; automatically translating this to the new name %r"%(handler, new_handler),
                  DeprecationWarning)
        return new_handler
    else:
        return handler

def get_named_handler_class(handler):
    """Get a HandlerBase identified by the given name.

    The name is of the format submodule.ClassName

    :rtype HandlerBase"""
    handler = _map_deprecated_handler_name(handler)
    try:
        output_module = importlib.import_module('.'+handler.split('.')[0],__name__)
    except ImportError:
        output_module = importlib.import_module(handler.split('.')[0])
    output_class = getattr(output_module, handler.split('.')[1])
    return output_class



"""Logic to help flexibly get/set data in ORM objects (specifically HaloProperty and SimulationProperty) where data
of different types has to be stored in different attributes.
"""

import datetime
import functools
import pickle
import sys
import time
import zlib

import numpy as np

pickle_loads = pickle.loads
if int(sys.version[0])==3:
    pickle_loads = functools.partial(pickle.loads, encoding='latin1')


_THRESHOLD_FOR_COMPRESSION = 1000

def get_data_of_unknown_type(obj):
    """Starting from the ORM object, extract data which may be stored in a variety of attributes depending on its type"""
    mapper = DataAttributeMapper(db_object=obj)
    return mapper.get(obj)

def set_data_of_unknown_type(obj, data):
    """Store the given data in the ORM object, using an appropriate attribute for the given data type"""
    mapper = DataAttributeMapper(data=data)
    mapper.set(obj,data)


class DataAttributeMapper:
    _order = 0
    # this can be used to force a subclass to be 'found' last
    # see __all_subclasses below, and use in NullAttributeMapper
    #
    # Except for a particular branch of py 3.5 it seems that __subclasses__ always
    # returns in order that classes were defined anyway, so it is not clear why this
    # is necessary.

    def __new__(cls, db_object=None, data=None):
        if db_object is None and data is None:
            subclass=NullAttributeMapper
        if data is None:
            subclass=cls._subclass_from_db_object(db_object)
        else:
            subclass=cls._subclass_from_data(data)

        return object.__new__(subclass)

    @classmethod
    def __all_subclasses(cls):
        subclasses = cls.__subclasses__()
        subclasses.sort(key = lambda x: x._order)
        for X in subclasses:
            yield X
            yield from X.__all_subclasses()

    @classmethod
    def __all_nonabstract_subclasses(cls):
        X= [sub for sub in cls.__all_subclasses() if hasattr(sub,"_attribute_name")]
        return X

    @classmethod
    def _subclass_from_db_object(cls, db_object):
        for subclass in cls.__all_nonabstract_subclasses():
            if subclass._handles_db_object(db_object):
                return subclass

        # NullAttributeMapper should prevent the following line from being reached
        assert False, "Internal error in data_attribute_mapper: no subclass claimed responsibility for object"

    @classmethod
    def _subclass_from_data(cls, data):
        for subclass in cls.__all_nonabstract_subclasses():
            if subclass._handles_data(data):
                return subclass
        raise TypeError("Don't know how to store data of type %r"%type(data))

    @classmethod
    def _handles_db_object(cls, db_object):
        if cls._attribute_name is None:
            return True
        return getattr(db_object, cls._attribute_name, None) is not None

    @classmethod
    def _handles_data(cls, data):
        return any([isinstance(data, t) for t in cls._handled_types])

    def pack(self, data):
        return data

    def unpack(self,data):
        return data

    def _clear_other_attributes(self, db_object):
        for cls in DataAttributeMapper.__all_nonabstract_subclasses():
            if not isinstance(self, cls) and cls._attribute_name is not None:
                setattr(db_object, cls._attribute_name, None)

    def set(self, db_object, data):
        if hasattr(db_object,self._attribute_name):
            setattr(db_object, self._attribute_name, self.pack(data))
        else:
            raise TypeError("%r object does not have a slot for %r"%(type(db_object),self._attribute_name))

        self._clear_other_attributes(db_object)

    def get(self, db_object):
        return self.unpack(getattr(db_object, self._attribute_name))

class TimeAttributeMapper(DataAttributeMapper):
    _attribute_name = "data_time"
    _handled_types = [datetime.datetime, time.struct_time]

    def pack(self, data):
        if isinstance(data, time.struct_time):
            data = datetime.datetime(*data[:6])
        return data

class StringAttributeMapper(DataAttributeMapper):
    _attribute_name = "data_string"
    _handled_types = [str, str]

class ArrayDowncastingAttributeMapper(DataAttributeMapper):
    @classmethod
    def _handles_data(cls, data):
        if any([isinstance(data, t) for t in cls._handled_types]):
            return True
        if isinstance(data, list) and len(data)==1 and cls._handles_data(data[0]):
            return True
        if isinstance(data, np.ndarray) and len(data.shape)==0 and \
            any([data.dtype == t for t in cls._handled_types]):
            return True
        return False

    def pack(self, data):
        if isinstance(data,list):
            return data[0]
        else:
            return self._handled_types[0](data)

class FloatAttributeMapper(ArrayDowncastingAttributeMapper):
    _attribute_name = "data_float"
    _handled_types = [float, np.float32, np.float64]
    if hasattr(np, 'float128'):
        _handled_types.append(np.float128)


class IntAttributeMapper(ArrayDowncastingAttributeMapper):
    _attribute_name = "data_int"
    _handled_types = [int, np.int32, np.int64, np.uint32, np.uint64]


class ArrayAttributeMapper(DataAttributeMapper):
    _attribute_name = "data_array"
    _handled_types = [list, np.ndarray]
    _order = 1 # must be used only when downcasting mappers have failed

    def _unpack_compressed(self, packed):
        return pickle_loads(zlib.decompress(packed[2:]))

    def _unpack_uncompressed(self, packed):
        return pickle_loads(packed[2:])

    def _unpack_old_format(self, packed):
        return np.frombuffer(packed)

    def unpack(self, packed):
        if len(packed)==0:
            return None
        elif packed.startswith(b"ZX"):
            return self._unpack_compressed(packed)
        elif packed.startswith(b"PX"):
            return self._unpack_uncompressed(packed)
        else:
            return self._unpack_old_format(packed)

    def pack(self, data):
        dumped_st = pickle.dumps(data)
        if len(dumped_st) > _THRESHOLD_FOR_COMPRESSION:
            dumped_st = b"ZX" + zlib.compress(dumped_st)
        else:
            dumped_st = b"PX" + dumped_st
        return dumped_st


# Following must be defined last to act as a fall-through:
class NullAttributeMapper(DataAttributeMapper):
    _handled_types = [type(None)]
    _attribute_name = None
    _order = 2 # see discussion in the DataAttributeMapper

    def set(self, db_object, data):
        assert data is None
        self._clear_other_attributes(db_object)

    def get(self, db_object):
        return None

__all__ = ['get_data_of_unknown_type', 'set_data_of_unknown_type']



from sqlalchemy import and_, func, select


def argmax(query, maximise_column, group_bys):
    """Returns a modified query that emulates an argmax function for SQL

    :param query: The SQLAlchemy query
    :param maximise_column: The table column to maximise
    :param group_bys: The table columns to group by
    :returns: modified SQLAlchemy query that returns the rows which satisfy argmax(maximise_column)
    """

    # we're going to have a subquery that looks very much like the main query,
    # but which serves to pick out the argmax
    argmax_subquery = query

    argmax_subquery = argmax_subquery.with_entities(func.max(maximise_column).label("maximise_column_result"),
                                                    *group_bys). \
        group_by(*group_bys).subquery()
    # NB tried making the above a materialised CTE to improve performance, but it doesn't help (presumably the SQLite
    # optimizers already spot the subquery is used twice?)

    group_bys_in_argmax_subquery = [argmax_subquery.corresponding_column(group_by) for group_by in group_bys]

    join_conditions = [a == b for a, b in zip(group_bys, group_bys_in_argmax_subquery)]

    # Ideally now the main query will join to the subquery to pick out all details of the selected row, like:
    #
    # query = query.join(argmax_subquery,
    #                    and_(maximise_column == argmax_subquery.c.maximise_column_result,
    #                         *join_conditions))
    #
    # However this ignores the edge case where two rows in a group both have the same value in maximise_column.
    #
    # We need to de-dup these groups, and do it by picking out the first of any duplicates, in order of the
    # primary key

    assert len(group_bys[0].table.primary_key.columns) == 1 # unable to de-dup if there isn't a single PK column

    primary_key = group_bys[0].table.primary_key.columns[0]

    deduped_argmax_subquery \
        = query.with_entities(func.max(primary_key).label('id')).join(argmax_subquery,
                                                    and_( maximise_column == argmax_subquery.c.maximise_column_result,
                                                            *join_conditions) ).\
                                                    group_by(*group_bys_in_argmax_subquery).subquery()


    # the main query will join to the subquery to pick out all details of the selected row
    query = query.join(deduped_argmax_subquery, deduped_argmax_subquery.c.id == primary_key)

    return query

def delete_non_maximal_rows(connection, table, maximise_column, group_bys):
    """Deletes columns from a table that are *not* the argmax.

    This is provided as a performance workaround because of the high cost of the argmax query

    :param connection: connection on which to execute the delete
    :param table: The SQLAlchemy table
    :param maximise_column: The table column to maximise
    :param group_bys: The table columns to group by
    :returns: number of deleted rows
    """
    max_weight = func.max(maximise_column).label("max_weight")

    source_and_max_weight = select(max_weight, *group_bys).\
        group_by(*group_bys).subquery(name="max_from_source_id")

    group_bys_subquery = [source_and_max_weight.corresponding_column(c) for c in group_bys]
    ids_to_eliminate = select(table.primary_key.columns[0]).\
        join(source_and_max_weight,and_(*[a==b for a,b in zip(group_bys, group_bys_subquery)])).\
        filter(maximise_column < source_and_max_weight.c.max_weight)

    # with sqlite it's OK to just do in_(ids_to_eliminate) but MySQL needs a subquery construction, otherwise
    # it complains it can't do the delete
    ids_to_eliminate_cte = ids_to_eliminate.subquery()
    ids_to_eliminate_select = select(ids_to_eliminate_cte.c.id)

    delete_statement = table.delete().where(table.primary_key.columns[0].in_(ids_to_eliminate_select))
    deleted_count = connection.execute(delete_statement).rowcount

    return deleted_count



import numpy as np
from sqlalchemy import BigInteger, Column, ForeignKey, Integer, orm, types
from sqlalchemy.orm import Session, backref, relationship

from . import Base, creator, extraction_patterns
from .dictionary import get_dict_id, get_or_create_dictionary_item
from .timestep import TimeStep


class UnsignedInteger(types.TypeDecorator):
    """Stores an unsigned int64 as a signed int64"""

    impl = types.Integer

    cache_ok = True

    def process_bind_param(self, value, dialect):
        return np.uint64(value).astype(np.int64)

    def process_result_value(self, value, dialect):
        if value is not None:
            if type(value) is bytes:
                return np.frombuffer(value, dtype=np.uint64)[0]
            else:
                return np.int64(value).astype(np.uint64)
        else:
            return None

class SimulationObjectBase(Base):
    __tablename__= "halos"

    id = Column(Integer, primary_key=True) #the unique ID value of the database object created for this halo
    halo_number = Column(BigInteger) #by default this will be the halo's rank in terms of particle count
    finder_id = Column(BigInteger) #raw halo ID from the halo catalog
    finder_offset = Column(BigInteger) #index of halo within halo catalog, primary identifier used when reading catalog/simulation data
    timestep_id = Column(Integer, ForeignKey('timesteps.id'))
    timestep = relationship(TimeStep, backref=backref(
        'objects', order_by=halo_number, cascade_backrefs=False, lazy='dynamic'), cascade='')
    NDM = Column(Integer)
    NStar = Column(Integer)
    NGas = Column(Integer)
    creator = relationship(creator.Creator, backref=backref(
        'halos', cascade_backrefs=False, lazy='dynamic'), cascade='save-update')
    creator_id = Column(Integer, ForeignKey('creators.id'))
    object_typecode = Column(Integer, nullable=False,
                         name='halo_type') # name for backwards compatibility

    tag = "abstract_base_class_for_halos_etc" # tag will be halo, bh, tracker etc. Don't use the base class!

    __mapper_args__ = {
        'polymorphic_identity':100,
        'polymorphic_on':object_typecode
    }

    @classmethod
    def _all_subclasses(cls):
        for c in cls.__subclasses__():
            yield c
            yield from c._all_subclasses()

    @staticmethod
    def class_from_tag(match_tag):
        match_tag = match_tag.lower()
        for c in SimulationObjectBase._all_subclasses():
            if match_tag == c.tag.lower():
                return c
        raise ValueError("Unknown object type %r"%match_tag)

    @staticmethod
    def object_typecode_from_tag(match_tag):
        if isinstance(match_tag, str):
            return SimulationObjectBase.class_from_tag(match_tag).__mapper_args__['polymorphic_identity']
        else:
            return match_tag

    @staticmethod
    def object_typetag_from_code(typecode):
        for c in SimulationObjectBase._all_subclasses():
            if c.__mapper_args__['polymorphic_identity'] == typecode:
                return c.tag
        raise ValueError("Unknown object typecode %d",typecode)

    @staticmethod
    def typecode_and_number_from_human_identifier(identifier):
        """Return object typecode and halo number given a human identification string or number

        E.g. identifier = 5 -> typecode=0,
        identifier = '1.5' -> typecode=1, number = 5
        identifier = 'bh_5' -> typecode=1, number = 5
        """
        if isinstance(identifier, int):
            object_typecode, object_number = 0, identifier
        elif "." in identifier:
            object_typecode, object_number = list(map(int, identifier.split(".")))
        elif "_" in identifier:
            object_typecode, object_number = identifier.split("_")
            object_typecode = SimulationObjectBase.class_from_tag(object_typecode).__mapper_args__['polymorphic_identity']
            object_number = int(object_number)
        else:
            object_typecode, object_number = 0, int(identifier)

        return object_typecode, object_number

    def __init__(self, timestep, halo_number, finder_id, finder_offset, NDM, NStar, NGas, object_typecode=None):
        self.timestep = timestep
        self.halo_number = int(halo_number)
        self.finder_id = int(finder_id)
        self.finder_offset = int(finder_offset)
        self.NDM = int(NDM)
        self.NStar = int(NStar)
        self.NGas = int(NGas)
        if object_typecode is not None:
            self.object_typecode = int(object_typecode)
        self.init_on_load()
        self.creator = creator.get_creator(Session.object_session(timestep))

    @orm.reconstructor
    def init_on_load(self):
        self._dict_is_complete = False
        self._d = {}

    def __repr__(self):

        return "<%s %r | NDM=%d Nstar=%d Ngas=%d>"%(self.__class__.__name__, self.path, self.NDM, self. NStar, self.NGas)


    @property
    def basename(self):
        return self.tag+"_"+str(self.halo_number)

    @property
    def path(self):
        return self.timestep.path+"/"+self.basename

    @property
    def handler(self):
        if not hasattr(self, "_handler"):
            self._handler = self.timestep.simulation.get_output_handler()
        return self._handler

    @property
    def handler_class(self):
        if not hasattr(self, "_handler_class"):
            self._handler_class = self.timestep.simulation.output_handler_class
        return self._handler_class

    def load(self, mode=None):
        """Load the data for this halo, if it is present on this computer's filesystem.

        :param mode: the load mode to pass to the relevant input handler. For example with the pynbody input
        handler this can be None or 'partial' (in a normal session) and, when running inside an MPI session,
        'server' or 'server-partial'. See https://pynbody.github.io/tangos/mpi.html.
        """
        halo_number = self.halo_number
        if not hasattr(self, "finder_id"):
            finder_id = self.halo_number # backward compatibility
        else:
            finder_id = self.finder_id
        if not hasattr(self, "finder_offset"):
            finder_offset = finder_id
        else:
            finder_offset = self.finder_offset

        return self.handler.load_object(self.timestep.extension, finder_id, finder_offset, object_typetag=self.tag, mode=mode)

    def calculate(self, calculation, return_description=False):
        """Use the live-calculation system to calculate a user-specified function of the stored data.

        See the data exploration tutorials at https://pynbody.github.io/tangos/data_exploration.html
        for an introduction to the system.

        :param calculation: the calculation (or a string representation of it to be parsed)
        :param return_description: if True, return both the value and the PropertyCalculation class describing it.
        :returns: The result of the calculation, or a tuple containing the result and the description if
                  return_description is True.

        """

        from .. import live_calculation
        calculation = live_calculation.parser.parse_property_name_if_required(calculation)
        (value,), description = calculation.values_sanitized_and_description([self], Session.object_session(self))
        if len(value)==1:
            retval = value[0]
        else:
            retval = value

        if return_description:
            return retval, description
        else:
            return retval

    def __getitem__(self, key):
        """Highest-level method for retrieving data or link"""
        return self.get_data(key)

    def get(self, key, default=None):
        """Highest-level method for retrieving data or link, with default return value if no data is found
        under the specified key"""
        try:
            return self.get_data(key)
        except KeyError:
            return default

    def get_data(self, key, raw=False, always_return_array=False):
        """High-level data access to the property or linked halo named by key.

        :param key: string with the name of the property or link
        :param raw: if True, get the raw data rather than attempt to reassemble the data into a science-ready form
        :param always_return_array: if True, always return the data as a list of values, even if there is only one
           stored property with the specified name
        """

        if raw:
            getters=[extraction_patterns.HaloPropertyRawValueGetter()]
        else:
            getters=[extraction_patterns.HaloPropertyValueGetter()]
        getters+=[extraction_patterns.HaloLinkTargetGetter()]

        return_data = self.get_objects(key, getters)

        if (not always_return_array) and len(return_data) == 1:
            return_data = return_data[0]

        return return_data

    def get_objects(self, key, getters=None):
        """Get objects belonging to this halo named by the specified key.

        Compared to get_data, this allows access to the underlying HaloProperty or HaloLink objects, or to perform
        custom processing of the data by specifying particular extraction patterns to getters. For more information,
        see halo_data_extraction_patterns."""
        if getters is None:
            getters = [extraction_patterns.HaloPropertyGetter(), extraction_patterns.HaloLinkGetter()]
        from . import Session
        session = Session.object_session(self)
        key_id = get_dict_id(key, session=session)

        ret_values = []
        for g in getters:
            ret_values += g.get(self, key_id, session)

        if len(ret_values) == 0:
            raise KeyError("No such property %r" % key)
        return ret_values

    def get_description(self, key, getters=None):
        """Get a description of a named property or link, in the form of an object capable of calculating it.

        This can be helpful to extract meta-data such as the size of an image or steps of an array."""
        if getters is None:
            getters = [extraction_patterns.HaloPropertyGetter(), extraction_patterns.HaloLinkGetter()]
        object = self.get_objects(key, getters)[0]
        return object.description


    def __setitem__(self, key, obj):
        if isinstance(obj, SimulationObjectBase):
            self._setitem_one_halo(key, obj)
        elif hasattr(obj, '__len__') and all([isinstance(x,SimulationObjectBase) for x in obj]):
            self._setitem_multiple_halos(key, obj)
        else:
            self._setitem_property(key, obj)

    def _setitem_property(self, key, obj):
        from . import Session
        from .halo_data import HaloProperty

        session = Session.object_session(self)
        key = get_or_create_dictionary_item(session, key)
        X = self.properties.filter_by(name_id=key.id).first()
        if X is not None:
            X.data = obj
            X.creator = creator.get_creator(session)
        else:
            X = HaloProperty(self, key, obj)
            X.creator = creator.get_creator(session)
            session.add(X)
        session.commit()

    def _setitem_one_halo(self, key, obj):
        from . import HaloLink, Session
        session = Session.object_session(self)
        key = get_or_create_dictionary_item(session, key)
        X = self.links.filter_by(halo_from_id=self.id, relation_id=key.id).first()
        if X is None:
            X = HaloLink(self, obj, key)
            X.creator = creator.get_creator(session)
            session.add(X)
        else:
            X.halo_to = obj
            X.creator = creator.get_creator(session)
        session.commit()

    def _setitem_multiple_halos(self, key, obj):
        from . import Session
        from .halo_data import HaloLink

        session = Session.object_session(self)
        key = get_or_create_dictionary_item(session, key)
        self.links.filter_by(halo_from_id=self.id, relation_id=key.id).delete()
        links = [HaloLink(self, halo_to, key) for halo_to in obj]
        session.add_all(links)
        session.commit()


    def keys(self, getters = None):
        if getters is None:
            getters = [extraction_patterns.HaloPropertyGetter(), extraction_patterns.HaloLinkGetter()]
        from . import Session
        names = []
        session = Session.object_session(self)
        for g in getters:
            names+=g.keys(self, session)

        return names

    def __contains__(self, item):
        return item in list(self.keys())

    @property
    def earliest(self):
        if not hasattr(self, '_earliest'):
            from .. import relation_finding
            strategy = relation_finding.MultiHopMajorProgenitorsStrategy(self,order_by=['time_asc'],include_startpoint=True)
            self._earliest=strategy.first()

        return self._earliest

    @property
    def latest(self):
        if not hasattr(self, '_latest'):
            from .. import relation_finding
            strategy = relation_finding.MultiHopMajorDescendantsStrategy(self,order_by=['time_desc'],include_startpoint=True)
            self._latest=strategy.first()

        return self._latest

    def plot(self, name, *args, **kwargs):
        from . import Session
        name_id = get_dict_id(name, Session.object_session(self))
        data = self.properties.filter_by(name_id=name_id).first()
        return data.plot(*args, **kwargs)


    def calculate_for_descendants(self, *plist, **kwargs):
        """Run the specified calculations on this halo and its descendants

        Each argument is a string (or an instance of live_calculation.Calculation), following the syntax
        described in live_calculation.md.

        *kwargs*:

        :param nmax: The maximum number of descendants to consider (default 1000)
        :param strategy: The class to use to find the descendants (default relation_finding.MultiHopMajorDescendantsStrategy)
        """
        from .. import live_calculation
        from .. import query as db_query
        from .. import relation_finding
        from .. import temporary_halolist as thl
        from . import Session

        nmax = kwargs.get('nmax',1000)
        strategy = kwargs.get('strategy', relation_finding.MultiHopMajorDescendantsStrategy)
        strategy_kwargs = kwargs.get('strategy_kwargs', {})

        if isinstance(plist[0], live_calculation.Calculation):
            property_description = plist[0]
        else:
            property_description = live_calculation.parser.parse_property_names(*plist)

        # must be performed in its own session as we intentionally load in a lot of
        # objects with incomplete lazy-loaded properties
        session = Session()
        try:
            with strategy(db_query.get_halo(self.id, session), nhops_max=nmax,
                          include_startpoint=True, **strategy_kwargs).temp_table() as tt:
                raw_query = thl.halo_query(tt)
                query = property_description.supplement_halo_query(raw_query)
                results = query.all()
                return property_description.values_sanitized(results, Session.object_session(self))
        finally:
            session.close()

    def calculate_for_progenitors(self, *plist, **kwargs):
        """Run the specified calculations on the progenitors of this halo

        For more information see calculate_for_descendants.
        """
        from .. import relation_finding
        kwargs['strategy'] = relation_finding.MultiHopMajorProgenitorsStrategy
        return self.calculate_for_descendants(*plist, **kwargs)

    def reverse_property_cascade(self, *args, **kwargs):
        """The old alias for calculate_for_progenitors, retained for compatibility"""
        return self.calculate_for_progenitors(*args, **kwargs)

    def property_cascade(self, *args, **kwargs):
        """The old alias for calculate_for_descendants, retained for compatibility"""
        return self.calculate_for_descendants(*args, **kwargs)


    @property
    def next(self):
        if not hasattr(self, '_next'):
            from .. import relation_finding
            strategy = relation_finding.HopMajorDescendantStrategy(self)
            self._next=strategy.first()

        return self._next

    @property
    def previous(self):
        if not hasattr(self, '_previous'):
            from .. import relation_finding
            strategy = relation_finding.HopMajorProgenitorStrategy(self)
            self._previous=strategy.first()

        return self._previous

    def short(self):
        return "<Halo " + str(self.halo_number) + " of ...>"

class Halo(SimulationObjectBase):
    __mapper_args__ = {
        'polymorphic_identity': 0
    }

    tag = "halo"

class Tracker(SimulationObjectBase):
    __mapper_args__ = {
        'polymorphic_identity':3
    }

    tag = "tracker"

    def __init__(self, timestep, halo_number):
        super().__init__(timestep, halo_number, halo_number, halo_number, 0,0,0,
                                 self.__mapper_args__['polymorphic_identity'])

    @property
    def tracker(self):
        return self.timestep.simulation.trackers.filter_by(halo_number=self.halo_number).first()

    def load(self, mode=None):
        handler = self.timestep.simulation.get_output_handler()
        return handler.load_tracked_region(self.timestep.extension, self.tracker, mode=mode)


class BH(Tracker):
    __mapper_args__ = {
        'polymorphic_identity': 1
    }

    tag = "BH"


class Group(SimulationObjectBase):
    __mapper_args__ = {
        'polymorphic_identity':2
    }

    tag = "group"

    def __init__(self, *args):
        super().__init__(*args)
        self.object_typecode = 2




class PhantomHalo(SimulationObjectBase):
    __mapper_args__ = {
        'polymorphic_identity': 4
    }

    tag = "phantom"

    def __init__(self, timestep, halo_number, finder_id):
        super().__init__(timestep, halo_number, finder_id, finder_id, 0,0,0,
                                 self.__mapper_args__['polymorphic_identity'])


TimeStep.halos = orm.relationship(Halo, lazy='dynamic',viewonly=True,
                                  order_by=SimulationObjectBase.halo_number)
TimeStep.trackers = orm.relationship(Tracker,lazy='dynamic',viewonly=True,
                                  order_by=Tracker.halo_number)
TimeStep.bhs = orm.relationship(BH, lazy='dynamic',viewonly=True,
                                  order_by=BH.halo_number)
TimeStep.groups = orm.relationship(Group, lazy='dynamic',viewonly=True,
                                  order_by=Group.halo_number)
TimeStep.phantoms = orm.relationship(PhantomHalo, lazy='dynamic',viewonly=True,
                                  order_by=PhantomHalo.halo_number)



def register_postgresql_adapters():
    import numpy as np
    from psycopg2.extensions import AsIs, register_adapter

    asis_adapter = lambda x : AsIs(x)
    register_adapter(np.float64, asis_adapter)
    register_adapter(np.int64, asis_adapter)
    register_adapter(np.float32, asis_adapter)
    register_adapter(np.int32, asis_adapter)



"""Helper classes for defining translations between .stat file of different formats."""

import numpy as np


class Function:
    """Define a column which is actually a function of other columns"""
    def __init__(self, fn, *input_arg_names):
        self.fn = fn
        self.input_arg_names = input_arg_names

    def __call__(self, raw_input_names, raw_input_values):
        input_args = [raw_input_values[raw_input_names.index(name)] for name in self.input_arg_names]
        return self.fn(*input_args)

    def inputs(self):
        return self.input_arg_names

class Rename:
    """Define a column by renaming an existing column"""
    def __init__(self, name):
        self.name = name

    def __call__(self, raw_input_names, raw_input_values):
        return raw_input_values[raw_input_names.index(self.name)]

    def inputs(self):
        return [self.name]

class Value:
    """Define a column by a fixed value"""
    def __init__(self, value):
        self.value = value

    def __call__(self, raw_input_names, raw_input_values):
        return self.value

    def inputs(self):
        return []

class DefaultValue:
    """Give a default value to a column, in case it is absent"""
    def __init__(self, name, default_value):
        self.name = name
        self.default_value = default_value

    def __call__(self, raw_input_names, raw_input_values):
        val = raw_input_values[raw_input_names.index(self.name)]
        if val is None:
            return self.default_value
        else:
            return val

    def inputs(self):
        return [self.name]



import glob

import numpy as np

from .. import config, core, parallel_tasks, tracking
from ..input_handlers.changa_bh import BlackHolesLog, ShortenedOrbitLog
from ..log import logger
from ..util.check_deleted import check_deleted
from . import GenericTangosTool


class ChangaBHImporter(GenericTangosTool):
    tool_name = 'import-changa-bh'
    tool_description = 'Import black holes and link them to parent halos'

    @classmethod
    def add_parser_arguments(cls, parser):
        parser.add_argument('--sims', '--for', action='store', nargs=1,
                            metavar='simulation_name',
                            help='Specify a simulation to run on')
        parser.add_argument('--link-only', action='store_true',
                            help='Do not look for any new black holes; only link already-known black holes to new timesteps or halos')
        parser.add_argument('--backwards', action='store_true',
                            help="Scan through timesteps from low to high redshift (default is high to low)")

    def process_options(self, options):
        self.args = options

    def run_calculation_loop(self):
        parallel_tasks.database.synchronize_creator_object()
        self._session = core.get_default_session()

        if not self.args.link_only:
            self._import_black_holes()

        self._timelink_black_holes()

    def _get_timesteps(self):
        sim_query = core.sim_query_from_name_list(self.args.sims, self._session)
        timesteps = core.get_default_session().query(core.timestep.TimeStep).filter(
            core.timestep.TimeStep.simulation_id.in_([q.id for q in sim_query.all()])). \
            order_by(core.timestep.TimeStep.time_gyr).all()
        if self.args.backwards:
            timesteps = timesteps[::-1]
        return timesteps

    def _timelink_black_holes(self):
        query = core.sim_query_from_name_list(self.args.sims, self._session)
        for sim in query.all():
            pairs = list(zip(sim.timesteps[:-1], sim.timesteps[1:]))
            candidate_filenames = glob.glob(config.base + "/" + sim.basename + "/*.BHmergers")
            if len(candidate_filenames) == 0:
                logger.error("No merger file for " + sim.basename)
                return
            elif len(candidate_filenames) > 1:
                logger.error("Can't work out which is the merger file for " + sim.basename)
                logger.error("Found: %s", candidate_filenames)
                return
            self._bhmerger_filename = candidate_filenames[0]
            with self._session.no_autoflush:
                self._generate_halolinks(pairs)

    def _generate_halolinks(self, pairs):
        for ts1, ts2 in parallel_tasks.distributed(pairs):
            if BlackHolesLog.can_load(ts2.filename):
                bh_log = BlackHolesLog(ts2.filename)
            elif ShortenedOrbitLog.can_load(ts2.filename):
                bh_log = ShortenedOrbitLog(ts2.filename)
            else:
                logger.error("Warning! No orbit file found!")

            links = []
            mergers_links = []
            bh_map = {}
            logger.info("Gathering BH tracking information for steps %r and %r", ts1, ts2)
            with parallel_tasks.ExclusiveLock("bh"):
                dict_obj = core.get_or_create_dictionary_item(self._session, "tracker")
                dict_obj_next = core.get_or_create_dictionary_item(self._session, "BH_merger_next")
                dict_obj_prev = core.get_or_create_dictionary_item(self._session, "BH_merger_prev")

            track_links_n, idf_n, idt_n = tracking.get_tracker_links(self._session, dict_obj_next)
            bh_objects_1, nums1, id1 = self._get_bh_objs_numbers_and_dbids(ts1)
            bh_objects_2, nums2, id2 = self._get_bh_objs_numbers_and_dbids(ts2)
            tracker_links, idf, idt = tracking.get_tracker_links(self._session, dict_obj)

            idf_n = np.array(idf_n)
            idt_n = np.array(idt_n)

            if len(nums1) == 0 or len(nums2) == 0:
                logger.info("No BHs found in either step %r or %r... moving on", ts1, ts2)
                continue

            logger.info("Generating BH tracker links between steps %r and %r", ts1, ts2)
            o1 = np.where(np.in1d(nums1, nums2))[0]
            o2 = np.where(np.in1d(nums2, nums1))[0]
            if len(o1) == 0 or len(o2) == 0:
                continue
            with self._session.no_autoflush:
                for ii, jj in zip(o1, o2):
                    if nums1[ii] != nums2[jj]:
                        raise RuntimeError("BH iords are mismatched")
                    exists = np.where((idf == id1[ii]) & (idt == id2[jj]))[0]
                    if len(exists) == 0:
                        links.append(core.halo_data.HaloLink(bh_objects_1[ii], bh_objects_2[jj], dict_obj, 1.0))
                        links.append(core.halo_data.HaloLink(bh_objects_2[jj], bh_objects_1[ii], dict_obj, 1.0))
            logger.info("Generated %d tracker links between steps %r and %r", len(links), ts1, ts2)

            logger.info("Generating BH Merger information for steps %r and %r", ts1, ts2)
            for l in open(self._bhmerger_filename):
                l_split = l.split()
                t = float(l_split[6])
                bh_dest_id = int(l_split[0])
                bh_src_id = int(l_split[1])
                ratio = float(l_split[4])

                # ratios in merger file are ambiguous (since major progenitor may be "source" rather than "destination")
                # re-establish using the log file:
                try:
                    ratio = bh_log.determine_merger_ratio(bh_src_id, bh_dest_id)
                except (ValueError, AttributeError) as e:
                    logger.debug(
                        "Could not calculate merger ratio for %d->%d from the BH log; assuming the .BHmergers-asserted value is accurate",
                        bh_src_id, bh_dest_id)

                if t > ts1.time_gyr and t <= ts2.time_gyr:
                    bh_map[bh_src_id] = (bh_dest_id, ratio)

            self._resolve_multiple_mergers(bh_map)
            logger.info("Gathering BH merger links for steps %r and %r", ts1, ts2)
            with self._session.no_autoflush:
                for src, (dest, ratio) in bh_map.items():
                    if src not in nums1 or dest not in nums2:
                        logger.warning("Can't link BH %r -> %r; missing BH objects in database", src, dest)
                        continue
                    bh_src_before = bh_objects_1[nums1.index(src)]
                    bh_dest_after = bh_objects_2[nums2.index(dest)]

                    if ((idf_n == bh_src_before.id) & (idt_n == bh_dest_after.id)).sum() == 0:
                        mergers_links.append(
                            core.halo_data.HaloLink(bh_src_before, bh_dest_after, dict_obj_next, 1.0))
                        mergers_links.append(
                            core.halo_data.HaloLink(bh_dest_after, bh_src_before, dict_obj_prev, ratio))

            logger.info("Generated %d BH merger links for steps %r and %r", len(mergers_links), ts1, ts2)

            with parallel_tasks.ExclusiveLock("bh"):
                logger.info("Committing total %d BH links for steps %r and %r", len(mergers_links) + len(links), ts1,
                            ts2)
                self._session.add_all(links)
                self._session.add_all(mergers_links)
                self._session.commit()
                logger.info("Finished committing BH links for steps %r and %r", ts1, ts2)

    def _generate_missing_bh_objects(self, bh_iord, f, existing_obj_num):
        halo = []
        for bhi in bh_iord:
            if bhi not in existing_obj_num:
                halo.append(core.halo.BH(f, int(bhi)))

        return halo

    def _collect_bh_trackers(self, bh_iord, sim, existing_trackers):
        track = []
        for bhi in bh_iord:
            bhi = int(bhi)
            et = np.where(existing_trackers == bhi)[0]
            if len(et) == 0:
                tx = core.tracking.TrackData(sim, bhi)
                tx.particles = [bhi]
                tx.use_iord = True
                track.append(tx)
        return track

    def _get_bh_halo_assignments(self, pynbody_snapshot):
        pynbody_halos = pynbody_snapshot.halos()
        bh_cen_halos = None
        bh_halos = None
        import pynbody

        if isinstance(pynbody_halos, pynbody.halo.RockstarIntermediateCatalogue):
            bh_cen_halos = pynbody_halos.get_group_array(family='BH')
        elif isinstance(pynbody_halos, pynbody.halo.AHFCatalogue):
            bh_cen_halos = pynbody_halos.get_group_array(top_level=False, family='bh')
            bh_halos = pynbody_halos.get_group_array(top_level=True, family='bh')
        else:
            pynbody_snapshot['gp'] = pynbody_halos.get_group_array()
            bh_cen_halos = pynbody_snapshot.star['gp'][np.where(pynbody_snapshot.star['tform'] < 0)[0]]

        with check_deleted(pynbody_halos):
            del pynbody_halos

        return bh_cen_halos, bh_halos

    def _import_black_holes(self):
        for timestep in parallel_tasks.distributed(self._get_timesteps()):
            logger.info("Processing %s", timestep)

            try:
                timestep_particle_data = timestep.load()
            except:
                logger.warning("File not found - continuing")
                continue

            if len(timestep_particle_data.star) < 1:
                logger.warning("No stars - continuing")
                continue

            timestep_particle_data.physical_units()

            logger.info("Gathering existing BH halo information from database for step %r", timestep)

            bhobjs = timestep.bhs.all()
            existing_bh_nums = [x.halo_number for x in bhobjs]

            logger.info("...found %d existing BHs", len(existing_bh_nums))

            logger.info("Gathering BH info from simulation for step %r", timestep)
            bh_iord_this_timestep = timestep_particle_data.star['iord'][
                np.where(timestep_particle_data.star['tform'] < 0)[0]]
            bh_mass_this_timestep = timestep_particle_data.star['mass'][
                np.where(timestep_particle_data.star['tform'] < 0)[0]]

            logger.info("Found %d black holes for %r", len(bh_iord_this_timestep), timestep)

            logger.info("Updating BH trackdata and BH objects using on-disk information from %r", timestep)

            self._add_missing_trackdata_and_BH_objects(timestep, bh_iord_this_timestep, existing_bh_nums)
            self._session.expire_all()

            logger.info("Calculating halo associations for BHs in timestep %r", timestep)
            bh_cen_halos, bh_halos = self._get_bh_halo_assignments(timestep_particle_data)

            # re-order our information so that links refer to BHs in descending order of mass
            bh_order_by_mass = np.argsort(bh_mass_this_timestep)[::-1]
            bh_iord_this_timestep = bh_iord_this_timestep[bh_order_by_mass]
            if bh_halos is not None:
                bh_halos = bh_halos[bh_order_by_mass]
            if bh_cen_halos is not None:
                bh_cen_halos = bh_cen_halos[bh_order_by_mass]

            logger.info("Freeing the timestep particle data")
            with check_deleted(timestep_particle_data):
                del (timestep_particle_data)

            if bh_halos is not None:
                self._assign_bh_to_halos(bh_halos, bh_iord_this_timestep, timestep, "BH")
            if bh_cen_halos is not None:
                self._assign_bh_to_halos(bh_cen_halos, bh_iord_this_timestep, timestep, "BH_central", "host_halo")

    def _assign_bh_to_halos(self, bh_halo_assignment, bh_iord, timestep, linkname, hostname=None):
        linkname_dict_id = core.dictionary.get_or_create_dictionary_item(self._session, linkname)
        if hostname is not None:
            host_dict_id = core.dictionary.get_or_create_dictionary_item(self._session, hostname)
        else:
            host_dict_id = None

        logger.info("Gathering %s links for step %r", linkname, timestep)

        links, link_id_from, link_id_to = tracking.get_tracker_links(self._session, linkname_dict_id)
        halos = timestep.halos.filter_by(object_typecode=0).all()

        halo_nums = [h.halo_number for h in halos]
        halo_catind = [h.finder_offset for h in halos]
        halo_ids = np.array([h.id for h in halos])

        logger.info("Gathering bh halo information for %r", timestep)
        with parallel_tasks.lock.SharedLock("bh"):
            bh_database_object, existing_bh_nums, bhobj_ids = self._get_bh_objs_numbers_and_dbids(timestep)

        bh_links = []

        with self._session.no_autoflush:
            for bhi, haloi in zip(bh_iord, bh_halo_assignment):
                haloi = int(haloi)
                bhi = int(bhi)
                if haloi not in halo_catind:
                    logger.warning("Skipping BH in halo %d as no corresponding halo found in the database", haloi)
                    continue
                if bhi not in existing_bh_nums:
                    logger.warning("Can't find the database object for BH %d", bhi)
                    print(bhi)
                    print(existing_bh_nums)
                    continue

                bh_index_in_list = existing_bh_nums.index(bhi)
                halo_index_in_list = halo_catind.index(haloi)
                bh_obj = bh_database_object[bh_index_in_list]
                halo_obj = halos[halo_index_in_list]

                num_existing_links = ((link_id_from == halo_ids[halo_index_in_list]) & (
                            link_id_to == bhobj_ids[bh_index_in_list])).sum()
                if num_existing_links == 0:
                    bh_links.append(core.halo_data.HaloLink(halo_obj, bh_obj, linkname_dict_id))
                    if host_dict_id is not None:
                        bh_links.append(core.halo_data.HaloLink(bh_obj, halo_obj, host_dict_id))

        logger.info("Committing %d %s links for step %r...", len(bh_links), linkname, timestep)
        with parallel_tasks.ExclusiveLock("bh"):
            self._session.add_all(bh_links)
            self._session.commit()
        logger.info("...done")

    def _get_bh_objs_numbers_and_dbids(self, timestep):
        bh_database_object = timestep.bhs.all()
        existing_bh_nums = [x.halo_number for x in bh_database_object]
        bhobj_ids = np.array([x.id for x in bh_database_object])
        return bh_database_object, existing_bh_nums, bhobj_ids

    def _add_missing_trackdata_and_BH_objects(self, timestep, this_step_bh_iords, existing_bhobj_iords):
        with parallel_tasks.ExclusiveLock("bh"):
            track, track_nums = tracking.get_trackers(timestep.simulation)
            with self._session.no_autoflush:
                tracker_to_add = self._collect_bh_trackers(this_step_bh_iords, timestep.simulation, track_nums)
                halo_to_add = self._generate_missing_bh_objects(this_step_bh_iords, timestep, existing_bhobj_iords)

            self._session.add_all(tracker_to_add)
            self._session.add_all(halo_to_add)
            self._session.commit()
        logger.info("Committed %d new trackdata and %d new BH objects for %r", len(tracker_to_add),
                    len(halo_to_add), timestep)

    def _resolve_multiple_mergers(self, bh_map):
        for k, v in bh_map.items():
            if v[0] in bh_map:
                old_target = v[0]
                old_weight = v[1]
                bh_map[k] = bh_map[old_target][0], v[1] * bh_map[old_target][1]
                logger.info("Multi-merger detected; reassigning %d->%d (old) %d (new)", k, old_target, bh_map[k][0])
                self._resolve_multiple_mergers(bh_map)
                return



import multiprocessing
import multiprocessing.resource_tracker
import os
import signal
import sys
import threading
import time
from typing import Optional

import tblib.pickling_support

_slave = False
_rank = None
_size = None
_pipe = None
_recv_lock = None
_recv_buffer = []

_print_exceptions = True

# Compatibility fix for python >=3.8 on MacOS, where the default process start
# method changed:
if sys.version_info[:2]>=(3,3):
    mp_context = multiprocessing.get_context('fork')
else :
    mp_context = multiprocessing

class NoMatchingItem(Exception):
    pass


def send(data, destination, tag=0):
    _pipe.send((data, destination, tag))

def receive_any(source=None):
    return receive(source,None,True)


def receive(source=None, tag=0, return_tag=False):
    while True:
        try:
            item = _pop_first_match_from_reception_buffer(source, tag)
            if return_tag:
                return item
            else:
                return item[0]
        except NoMatchingItem:
            _receive_item_into_buffer()


NUMPY_SPECIAL_TAG = 1515

def send_numpy_array(data, destination):
    send(data,destination,tag=NUMPY_SPECIAL_TAG)

def receive_numpy_array(source):
    return receive(source,tag=NUMPY_SPECIAL_TAG)

def _pop_first_match_from_reception_buffer(source, tag):
    for item in _recv_buffer:
        if ((item[2] == tag or tag is None) and (item[1] == source or source is None)):
            # consume item
            _recv_buffer.remove(item)
            return item

    raise NoMatchingItem()

def _receive_item_into_buffer():
    if _recv_lock.acquire(False):
        try:
            _recv_buffer.append(_pipe.recv())
        finally:
            _recv_lock.release()
    else:
        # block until a data item has been received by another thread
        _recv_lock.acquire()
        _recv_lock.release()



def rank():
    return _rank

def size():
    return _size

def barrier():
    pass

def finalize():
    pass

def launch_wrapper(target_fn, rank_in, size_in, pipe_in, args_in, capture_log):
    tblib.pickling_support.install()

    global _slave, _rank, _size, _pipe, _recv_lock
    _rank = rank_in
    _size = size_in
    _pipe = pipe_in
    _recv_lock = threading.Lock()

    result = None

    try:
        if capture_log:
            from .. import log
            try:
                with log.LogCapturer() as lc:
                    target_fn(*args_in)
            finally:
                result = lc.get_output()
        else:
            target_fn(*args_in)
        if result is not None:
            _pipe.send(("log", result))
        _pipe.send("exit")
    except Exception as e:
        import sys
        import traceback
        exc_type, exc_value, exc_traceback = sys.exc_info()
        global _print_exceptions
        if _print_exceptions:
            print("Error on a sub-process:", file=sys.stderr)
            traceback.print_exception(exc_type, exc_value, exc_traceback,
                                      file=sys.stderr)
        if result is not None:
            _pipe.send(("log", result))
        _pipe.send(("error", exc_value, exc_traceback))


    _pipe.close()

class RemoteException(Exception):
    pass

def launch_functions(functions, args, capture_log=False):
    global _slave
    if _slave:
        raise RuntimeError("Multiprocessing session is already underway")

    # the resource tracker must be running before we start any processes,
    # otherwise they'll start their own resource trackers and all sorts
    # of confusion will ensue
    multiprocessing.resource_tracker.ensure_running()

    num_procs = len(functions)


    child_connections, parent_connections = list(zip(*[mp_context.Pipe() for rank in range(num_procs)]))
    processes = [mp_context.Process(target=launch_wrapper, args=(function, rank, num_procs, pipe, args_i, capture_log))
                 for rank, (pipe, function, args_i) in
                 enumerate(zip(child_connections, functions, args))]

    for proc_i in processes:
        proc_i.start()

    running = [True for rank in range(num_procs)]
    error: Optional[Exception] = None

    log = "" if capture_log else None

    while any(running):
        for i, pipe_i in enumerate(parent_connections):
            if pipe_i.poll():
                message = pipe_i.recv()
                if message=='exit':
                    running[i]=False
                elif isinstance(message[0], str) and message[0]=='error':
                    error = message[1]
                    traceback = message[2]
                    running = [False]
                    break
                elif isinstance(message[0], str) and message[0]=='log':
                    log+=message[1]
                else:
                    #print "multiprocessing backend: pass message ",i,"->",message[1]
                    parent_connections[message[1]].send((message[0],i,message[2]))

    #print "multiprocessing backend: all finished"

    for pipe_i in parent_connections:
        pipe_i.close()

    for proc_i in processes:
        if error:
            #print "multiprocessing backend: send signal to",proc_i.pid
            os.kill(proc_i.pid, signal.SIGTERM)
        proc_i.join()

    if error is not None:
        raise error.with_traceback(traceback)

    return _sort_log(log)

def _sort_log(log):
    """Sort the log by time.

    The input log is of the format, for example:
      [  3] 2023-12-12 13:55:46,004 Message

    This routine extracts the times, parses them and returns a reordered log in time order.
    It is stable, i.e. messages with the same timestep are returned in the input order
    """
    if log is None:
        return None
    lines = log.split("\n")
    times = []
    for line in lines:
        if line.strip() == "":
            continue
        times.append(line[6:28])
    times = [time.strptime(t, "%Y-%m-%d %H:%M:%S,%f") for t in times]
    lines = [line for _, line in sorted(zip(times, lines), key=lambda item: item[0])]
    return "\n".join(lines)





def launch(function, args, **kwargs):
    from .. import _num_procs
    if _num_procs is None:
        raise RuntimeError("To launch a parallel session using multiprocessing backend, you need to specify the number "
                           "of processors. You can do this by calling the backend multiprocessing-<n> where <n> is the"
                           "number of processors you want to use.")

    return launch_functions([function]*_num_procs, [args]*_num_procs, **kwargs)



import contextlib
import gc
import inspect
import os
import traceback

import sqlalchemy
import sqlalchemy.event
from sqlalchemy import create_engine, text

from .. import core, get_halo
from ..config import testing_db_backend, testing_db_password, testing_db_user


def _as_halos(hlist, session=None):
    if session is None:
        session = core.get_default_session()
    rvals = []
    for h in hlist:
        if h is None:
            rvals.append(None)
        elif isinstance(h, core.halo.SimulationObjectBase):
            rvals.append(h)
        elif isinstance(h, list) or isinstance(h, tuple):
            rvals.append(_as_halos(h, session))
        else:
            rvals.append(get_halo(h, session))
    return rvals

def _halos_to_strings(hlist):
    if len(hlist)==0:
        return "(empty list)"
    else:
        strlist = []
        for h in hlist:
            if isinstance(h, list) or isinstance(h, tuple):
                strlist.append(_halos_to_strings(h))
            elif isinstance(h,str):
                strlist.append(h)
            else:
                strlist.append(h.path)
        return str(strlist)

def halolists_equal(hl1, hl2, session=None):
    """Return True if hl1 and hl2 are equivalent lists of halos"""

    hl1 = _as_halos(hl1)
    hl2 = _as_halos(hl2)

    return len(hl1)==len(hl2) and all([h1==h2 for h1, h2 in zip(hl1,hl2)])

def assert_halolists_equal(hl1, hl2, session=None):
    equal = halolists_equal(hl1, hl2, session=None)
    assert equal, "Not equal: %s %s"%(_halos_to_strings(hl1),_halos_to_strings(hl2))

@contextlib.contextmanager
def autorevert():
    old_session = core.get_default_session()
    connection = core.get_default_engine().connect()
    transaction = connection.begin()
    isolated_session = core.Session(bind=connection)
    core.set_default_session(isolated_session)
    yield
    transaction.rollback()
    isolated_session.close()
    connection.close()
    core.set_default_session(old_session)

@contextlib.contextmanager
def assert_connections_all_closed():
    num_connections = [0,0]
    connection_details = {}
    def on_checkout(dbapi_conn, connection_rec, connection_proxy):
        num_connections[0]+=1
        num_connections[1]+=1
        connection_details[id(connection_rec)] = traceback.extract_stack()

    def on_checkin(dbapi_conn, connection_rec):
        if id(connection_rec) in connection_details:
            num_connections[0]-=1
            del connection_details[id(connection_rec)]

    gc.collect()

    sqlalchemy.event.listen(core.get_default_engine().pool, 'checkout', on_checkout)
    sqlalchemy.event.listen(core.get_default_engine().pool, 'checkin', on_checkin)

    yield


    gc.collect()

    sqlalchemy.event.remove(core.get_default_engine().pool, 'checkout', on_checkout)
    sqlalchemy.event.remove(core.get_default_engine().pool, 'checkin', on_checkin)

    for k,v in connection_details.items():
        print("object id",k,"not checked in; was created here:")
        for line in traceback.format_list(v):
            print("  ",line)

    assert num_connections[0]==0, "%d (of %d) connections were not closed"%(num_connections[0], num_connections[1])

class SqlExecutionTracker:
    """Logs queries performed against the given sqlalchemy connection.

    Based on https://stackoverflow.com/questions/19073099/how-to-count-sqlalchemy-queries-in-unit-tests

    Usage:
        with SqlExecutionCounter(conn) as ctr:
            conn.execute("SELECT 1")
            conn.execute("SELECT 1")
        assert ctr.count == 2
        assert "select" in ctr
        assert "update" not in ctr
    """
    def __init__(self, conn=None):
        if conn is None:
            conn = core.get_default_engine()

        self.conn = conn
        self._queries = []
        self._stacks = []

    def __enter__(self):
        sqlalchemy.event.listen(self.conn, 'after_execute', self.callback)
        return self

    def __exit__(self, *_):
        sqlalchemy.event.remove(self.conn, 'after_execute', self.callback)

    @property
    def count(self):
        return len(self._queries)

    def get_statement(self, i):
        return self._queries[i]

    def count_statements_containing(self, search_string):
        return sum(self.statements_contain(search_string))

    def traceback_statements_containing(self, search_string):
        return [tb for include, tb in zip(self.statements_contain(search_string),
                                          self._stacks)
                if include]

    def statements_contain(self, search_string):
        return [search_string.lower() in q.lower() for q in self._queries]

    def __contains__(self, search_string):
        return (any(self.statements_contain(search_string)))

    def callback(self, conn, query, *_):
        self._queries.append(str(query))
        self._stacks.append("".join(traceback.format_list(traceback.extract_stack()[:-2])))

def init_blank_db_for_testing(**init_kwargs):
    """Initialize a blank database for testing purposes.

    :arg testing_db_name: name of the database to create.  Defaults to the name of the calling file.
    :arg erase_if_exists: if True (default), delete the database if it already exists.

    Other arguments are passed to tangos.core.init_db

    :returns: True if a blank database has been created. False if the database already existed and erase_if_exists was False.
    """

    db_is_blank = False

    core.close_db()

    try:
        os.mkdir("test_dbs")
    except OSError:
        pass

    caller_fname = os.path.basename(inspect.getframeinfo(inspect.currentframe().f_back)[0])[:-3]

    testing_db_name = init_kwargs.pop("testing_db_name", caller_fname)
    erase_if_exists = init_kwargs.pop("erase_if_exists", True)

    if testing_db_backend == "sqlite":
        db_name = f"test_dbs/%s.db"%testing_db_name
        if erase_if_exists:
            try:
                os.remove(db_name)
            except OSError:
                pass
            db_is_blank = True
        else:
            db_is_blank = not os.path.exists(db_name)

        core.init_db(f"sqlite:///{db_name}", **init_kwargs)
    else:
        db_url = f"{testing_db_backend}://{testing_db_user}:{testing_db_password}@localhost"
        engine = create_engine(db_url)

        if erase_if_exists:
            with engine.connect() as conn:
                conn.execute(text("COMMIT"))
                conn.execute(text(f"DROP DATABASE IF EXISTS {testing_db_name}"))
                conn.execute(text("COMMIT"))
                conn.execute(text(f"CREATE DATABASE {testing_db_name}"))
            db_is_blank = True
        else:
            try:
                create_engine(f"{db_url}/{testing_db_name}").connect()
            except sqlalchemy.exc.OperationalError:
                with engine.connect() as conn:
                    conn.execute(text("COMMIT"))
                    conn.execute(text(f"CREATE DATABASE {testing_db_name}"))
                    db_is_blank = True
        core.init_db(f"{db_url}/{testing_db_name}", **init_kwargs)

    return db_is_blank



from . import YtPropertyCalculation


class GasMass(YtPropertyCalculation):
    """Calculates masses of total and cold (T<1.5x10^4 K) gas
    within a sphere of radius Rvir surrounding the center of a
    given halo"""
    names = "Mgas","Mcoldgas"

    def requires_property(self):
        return ["center", "Rvir_kpc"]

    def calculate(self, particle_data, existing_properties):
         cgas = particle_data.cut_region(["obj['temperature'] < 1.5e4"])
         return float(particle_data.quantities.total_quantity(('gas', 'cell_mass')).in_units('Msun').value),\
                float(cgas.quantities.total_quantity(('gas', 'cell_mass')).in_units('Msun').value)



import warnings


def send(data, destination, tag=0):
    raise RuntimeError("Cannot send data to another CPU: parallelism is disabled")

def receive(source=None, tag=0):
    raise RuntimeError("Cannot receive data from another CPU: parallelism is disabled")

def receive_any(source=None):
    raise RuntimeError("Cannot receive data from another CPU: parallelism is disabled")

def rank():
    return 0

def size():
    return 1

def barrier():
    pass

def finalize():
    pass


def launch(function, args):
    function(*args)



#!/usr/bin/env python


import sys


def run_importer(argv):
    from tangos import core, parallel_tasks
    from tangos.tools.property_importer import PropertyImporter
    importer = PropertyImporter()
    importer.parse_command_line(argv)
    parallel_tasks.launch(importer.run_calculation_loop,  [])

def main():
    print("""
       The 'tangos_import_from_ahf' command line is deprecated in favour of 'tangos property-import'.
       'tangos_import_from_ahf' may be removed in future versions.
       """)
    run_importer(sys.argv[1:])



from sqlalchemy import Column, ForeignKey, Integer
from sqlalchemy.orm import Session, backref, relationship

from ...config import DOUBLE_PRECISION
from .. import Base, creator
from ..dictionary import DictionaryItem
from ..halo import SimulationObjectBase


class HaloLink(Base):
    __tablename__ = 'halolink'

    id = Column(Integer, primary_key=True)
    halo_from_id = Column(Integer, ForeignKey('halos.id'))
    halo_to_id = Column(Integer, ForeignKey('halos.id'))

    halo_from = relationship(SimulationObjectBase, primaryjoin=halo_from_id == SimulationObjectBase.id,
                             backref=backref('links', cascade_backrefs=False,
                                             lazy='dynamic', viewonly=True,
                                             primaryjoin=halo_from_id == SimulationObjectBase.id),
                             cascade='')

    halo_to = relationship(SimulationObjectBase, primaryjoin=(halo_to_id == SimulationObjectBase.id),
                           backref=backref('reverse_links', cascade_backrefs=False,
                                           lazy='dynamic', viewonly=True,
                                           primaryjoin=halo_to_id == SimulationObjectBase.id),
                           cascade='')

    weight = Column(DOUBLE_PRECISION)

    creator_id = Column(Integer, ForeignKey('creators.id'))
    creator = relationship(creator.Creator, backref=backref(
        'halolinks', cascade_backrefs=False, lazy='dynamic'), cascade='save-update')

    relation_id = Column(Integer, ForeignKey('dictionary.id'))
    relation = relationship(DictionaryItem, primaryjoin=(
        relation_id == DictionaryItem.id), cascade='')



    def __init__(self,  halo_from, halo_to, relationship, weight=1.0):
        self.halo_from = halo_from
        self.halo_to = halo_to

        self.relation = relationship
        self.weight = weight

        self.creator = creator.get_creator(Session.object_session(halo_from))

    def __repr__(self):
        if self.weight is None:
            weight_str = "None"
        else:
            weight_str = "%.2f"%self.weight
        return "<HaloLink " + str(self.relation.text) + " " + str(self.halo_from.path) + " to " + str(self.halo_to.path) \
               + " weight=%s>"%weight_str



import contextlib
import functools

import numpy as np

from .. import LivePropertyCalculation
from . import PynbodyPropertyCalculation
from .spherical_region import SphericalRegionPropertyCalculation


class CentreAndRadius(PynbodyPropertyCalculation):
    names = "shrink_center", "max_radius"

    def calculate(self, halo, existing_properties):
        dm_center, dm_max_radius = self._get_centre_and_max_radius(halo.dm)
        return dm_center, dm_max_radius

    def _get_centre_and_max_radius(self, particle_data):
        # this does not appear at module level because we want tangos to be importable even when pynbody is not
        # present:
        import pynbody

        # Throw a more informative error message if particle data is empty
        if len(particle_data) == 0:
            raise RuntimeError("No particle of this type are available for centering")

        # ensure the box is wrapped correctly by centring on one of the particles:
        temporary_centre = np.array(particle_data['pos'][0])
        with _recenter(particle_data, temporary_centre):
            center = pynbody.analysis.halo.shrink_sphere_center(particle_data, shrink_factor=0.8, velocity=False)

            # mark_timer can be used to track timing of parts of the calculation. The results of these timings
            # appears in the tangos_writer logs:
            self.mark_timer('cen')

            if any(center != center):
                raise RuntimeError("Something bizarre has happened with the centering")

            # measure rmax from the robust centre we have now identified:
            particle_data['pos'] -= center

            center += temporary_centre  # centre should be relative to original snapshot!

            rmax = pynbody.derived.r(particle_data).max()
            # N.B. not using halo['r'] which could end up calculating r across entire simulation, needlessly

        return center.view(np.ndarray), rmax


class CentreAndRadiusStars(CentreAndRadius, SphericalRegionPropertyCalculation):
    names = "stars_shrink_center", "stars_max_radius"
    # Inherit from Spherical Region in case halo finder does not link gas
    # and star particles to the actual "halo" object
    # in pynbody (as for HOP)

    def calculate(self, halo, existing_properties):
        stars_center, stars_max_radius = self._get_centre_and_max_radius(halo.st)
        return stars_center, stars_max_radius


class CentreAndRadiusGas(CentreAndRadius, SphericalRegionPropertyCalculation):
    names = "gas_shrink_center", "gas_max_radius"

    def calculate(self, halo, existing_properties):
        gas_center, gas_max_radius = self._get_centre_and_max_radius(halo.g)
        return gas_center, gas_max_radius


class CentreAndRadiusComoving(LivePropertyCalculation):
    names = "shrink_center_comoving", "max_radius_comoving"

    def requires_property(self):
        return ['shrink_center','max_radius']

    def calculate(self, _, halo):
        scalefactor = 1./(1.+halo.timestep.redshift)
        return halo['shrink_center']/scalefactor, halo['max_radius']/scalefactor


@contextlib.contextmanager
def _recenter(halo, centre):
    original_positions = np.array(halo['pos'])  # take a copy so we can put everything back at the end
    halo['pos'] -= centre
    halo.wrap()
    yield
    halo['pos'] = original_positions


def centred_calculation(fn):
    """Wrap a calculation with a robust recentring of the halo particles that is automatically reverted"""
    @functools.wraps(fn)
    def new_fn(self, halo, existing_properties):
        # Note that we recenter halo.ancestor i.e. the whole snapshot, if it is loaded into memory, so that there
        # is no confusion when performing whole-snapshot operations such as smoothing/density calculations.
        with _recenter(halo.ancestor, existing_properties['shrink_center']):
            return fn(self, halo, existing_properties)

    return new_fn



"""Code to read Peter Behroozi's Rockstar/consistent-trees output"""

import os

import numpy as np

from ..log import logger


class ConsistentTrees:
    def __init__(self, path):
        self._path = self._infer_subpath(path)
        self._load_raw_trees()
        self._load_scale_to_snap_number()
        self._setup_map_to_original_finder_catalogues()

    def _infer_subpath(self, rootpath):
        if os.path.exists(os.path.join(rootpath,'trees')):
            return rootpath
        elif os.path.exists(os.path.join(rootpath, 'halos', 'trees')):
            return os.path.join(rootpath, 'halos')
        else:
            raise OSError("Cannot find the consistent-trees output")

    def _load_raw_trees(self):
        filename = os.path.join(self._path, "trees", "tree_0_0_0.dat")
        f = open(filename)
        while True:
            l = f.readline()
            if l.startswith("#tree"):
                break
        read_type = np.dtype([
                  ('id_this', np.int64),
                  ('id_desc', np.int64),
                  ('phantom', np.int8),
                  ('Mvir', np.float32)])
        results = np.loadtxt(f, usecols=(1,3,8,10), dtype=read_type)
        self.links = results

    def _load_scale_to_snap_number(self):
        filename = os.path.join(self._path, "outputs", "scales.txt")
        read_type = np.dtype([('snap_number', np.int32), ('scale', np.float32)])
        self.scale_to_snap = np.loadtxt(filename, dtype=read_type)
        self._snap_min = self.scale_to_snap['snap_number'].min()
        self._snap_max = self.scale_to_snap['snap_number'].max()


    def _load_original_catalogue(self, snapnum):
        filename = os.path.join(self._path, "outputs", "really_consistent_%d.list"%snapnum)
        f = open(filename)
        collist = f.readline().split()
        f.close()
        if 'Original_ID' in collist:
            read_cols = (0,collist.index('Original_ID'))
            return np.loadtxt(filename, dtype=np.int64, usecols=read_cols, unpack=True)
        else:
            raise OSError("Cannot identify Original_ID column in really_consistent_%d.list"%snapnum)

    def _setup_map_to_original_finder_catalogues(self):
        maxval = self.links['id_this'].max()
        self._id_to_finder_id = np.zeros(maxval + 1, dtype=np.int64)
        self._id_to_snap_num = np.zeros(maxval + 1, dtype=np.int32) - 1
        for snapnum in range(self._snap_min, self._snap_max + 1):
            ctid, original_id = self._load_original_catalogue(snapnum)
            self._id_to_finder_id[ctid] = original_id
            self._id_to_snap_num[ctid] = snapnum

        self._sanity_check_snap_num_assignment()
        self._snap_nums = self._get_snapshot_nums(self.links['id_this'])

        ids = self.links['id_this']
        phantom = self.links['phantom']
        # any remaining IDs correspond to phantom halos
        for snapnum in range(self._snap_min, self._snap_max + 1):
            phantom_mask = ((self._snap_nums == snapnum) & (phantom != 0))
            num_phantoms = phantom_mask.sum()
            phantom_ids = -np.arange(1, num_phantoms + 1)
            self._id_to_finder_id[ids[phantom_mask]] = phantom_ids

    def _sanity_check_snap_num_assignment(self):
        snap_nums_this = self._get_snapshot_nums(self.links['id_this'])
        snap_nums_next = self._get_snapshot_nums(self.links['id_desc'])
        snap_nums_next[snap_nums_next == -1] = self._snap_max + 1
        snap_nums_diff = snap_nums_next-snap_nums_this
        assert (snap_nums_diff == 1).all()

    def _get_finder_ids(self, consistent_trees_ids):
        return self._id_to_finder_id[consistent_trees_ids]

    def _get_snapshot_nums(self, consistent_trees_ids):
        snapnums = self._id_to_snap_num[consistent_trees_ids]
        snapnums[consistent_trees_ids<0]=-1
        return snapnums

    def get_num_phantoms_in_snapshot(self, snapnum):
        finder_ids = self._id_to_finder_id[self._id_to_snap_num==snapnum]
        if len(finder_ids)==0:
            return 0
        else:
            return -finder_ids.min()

    def get_links_for_snapshot(self, snapnum):
        """Get the links from snapshot snapnum to its immediate successor.

        Returns a dictionary; keys are the finder IDs at the given snapnum, whereas the values are
        a tuple containing the ID at the subsequent snapshot and the merger ratio (or 1.0 for no merger).

        Negative values indicate phantom halos, i.e. halos that were not present in the finder output"""

        this_snap_mask = self._snap_nums == snapnum
        ids_this_snap = self._id_to_finder_id[self.links['id_this'][this_snap_mask]]
        ids_next_snap = self._id_to_finder_id[self.links['id_desc'][this_snap_mask]]

        merger_ratios = self._get_merger_ratio_array(ids_next_snap, snapnum)

        return dict(zip(ids_this_snap, zip(ids_next_snap, merger_ratios)))

    def get_finder_id_to_tree_id_for_snapshot(self, snapnum):
        """Get the internal consistent-trees ids for each original halo-finder ID"""
        this_snap_mask = (self._snap_nums == snapnum)
        internal_ids_this_snap = self.links['id_this'][this_snap_mask]
        finder_ids_this_snap = self._id_to_finder_id[internal_ids_this_snap]
        finder_id_to_id = dict(zip(finder_ids_this_snap, internal_ids_this_snap))
        return finder_id_to_id

    def _get_merger_ratio_array(self, ids_next_snap, snapnum):
        this_snap_mask = self._snap_nums == snapnum
        ratio = np.ones(len(ids_next_snap))
        num_occurences_next_snap = np.bincount(ids_next_snap[ids_next_snap >= 0])
        mergers_next_snap = np.where(num_occurences_next_snap > 1)[0]
        logger.info("Identified %d mergers between snapshot %d and %d", len(mergers_next_snap), snapnum, snapnum + 1)
        for merger in mergers_next_snap:
            contributor_offsets = np.where(ids_next_snap == merger)[0]
            contributing_masses = self.links['Mvir'][this_snap_mask][contributor_offsets]
            ratio[contributor_offsets] = contributing_masses / contributing_masses.sum()
        return ratio



from . import core


def create_property(halo, name, prop, session):

    name = core.dictionary.get_or_create_dictionary_item(session, name)

    if isinstance(prop, core.halo.Halo):
        px = core.halo_data.HaloLink(halo, prop, name)
    else:
        px = core.halo_data.HaloProperty(halo, name, prop)

    px.creator = core.creator.get_creator(session)
    return px


def _insert_list_unlocked(property_list):
    session = core.get_default_session()
    number = 0
    for p in property_list:
        if p[2] is not None:
            session.add(create_property(p[0], p[1], p[2], session))
            number += 1

    session.commit()
    return number

def insert_list(property_list):
    from tangos import parallel_tasks as pt

    if pt.backend!=None:
        with pt.ExclusiveLock("insert_list"):
            return _insert_list_unlocked(property_list)
    else:
        return _insert_list_unlocked(property_list)



import numpy as np
import pynbody.array.shared

from ..message import Message


def send_array(array: pynbody.array.SimArray, destination: int, use_shared_memory: bool = False):
    if use_shared_memory:
        if not hasattr(array, "_shared_fname"):
            if isinstance(array, np.ndarray) and hasattr(array, "base") and hasattr(array.base, "_shared_fname"):
                array._shared_fname = array.base._shared_fname # the strides/offset will point into the same memory
            else:
                raise ValueError("Array %r has no shared memory information" % array)
        _send_array_shared_memory(array, destination)
    else:
        _send_array_copy(array, destination)

def receive_array(source: int, use_shared_memory: bool = False):
    if use_shared_memory:
        return _receive_array_shared_memory(source)
    else:
        return _receive_array_copy(source)

def _send_array_copy(array: np.ndarray, destination: int):
    from .. import backend
    backend.send_numpy_array(array, destination)

def _receive_array_copy(source):
    from .. import backend
    return backend.receive_numpy_array(source)


class SharedMemoryArrayInfo(Message):
    pass

def _send_array_shared_memory(array: pynbody.array.SimArray, destination: int):
    info = pynbody.array.shared._shared_array_deconstruct(array, transfer_ownership=False)
    SharedMemoryArrayInfo(info).send(destination)

def _receive_array_shared_memory(source):
    info = SharedMemoryArrayInfo.receive(source)
    return pynbody.array.shared._shared_array_reconstruct(info.contents)



import os


def read_datasets(basedir,filename):
    if os.path.exists(os.path.join(basedir, "datasets.txt")):
        with open(os.path.join(basedir, "datasets.txt")) as f:
            for l in f:
                if l.split()[0].endswith(filename):
                    return int(l.split()[1])
    else:
        raise AssertionError("Unable to open datasets.txt")



import time

from ..config import DEFAULT_SLEEP_BEFORE_ALLOWING_NEXT_LOCK
from . import log, message, parallelism_is_active


class MessageRequestLock(message.Message):
    def __init__(self, name, shared=False):
        self.name = name
        self.shared = shared

    @classmethod
    def deserialize(cls, source, message):
        obj = cls(*message)
        obj.source=source
        return obj

    def serialize(self):
        return (self.name, self.shared)

    def process(self):
        lock_id = self.name
        log.logger.debug("Received request for lock %r for proc %d, shared=%r", lock_id, self.source, self.shared)
        queue = _get_lock_queue(lock_id)
        queue.append((self.source, self.shared))
        if len(queue) == 1:
            _issue_next_lock(lock_id)
        elif _lock_in_shared_mode(lock_id) and self.shared:
            log.logger.debug("Issue shared lock %r to proc %d", lock_id, self.source)
            MessageGrantLock((lock_id, False)).send(self.source)
            _increment_lock_num_shared(lock_id,1)

class MessageRelinquishLock(message.Message):
    def process(self):
        lock_id = self.contents
        proc = self.source
        if _lock_in_shared_mode(lock_id):
            _release_lock_shared(lock_id, proc)
        else:
            _release_lock_exclusive(lock_id, proc)




class MessageGrantLock(message.Message):
    pass


_lock_queues = {}
_lock_num_sharers = {}

def _get_lock_queue(lock_id):
    lock_queue = _lock_queues.get(lock_id,[])
    _lock_queues[lock_id] = lock_queue
    return lock_queue

def _get_lock_num_sharers(lock_id):
    return _lock_num_sharers.get(lock_id, 0)

def _set_lock_num_sharers(lock_id, number):
    _lock_num_sharers[lock_id] = number

def _increment_lock_num_shared(lock_id, increment_by):
    _set_lock_num_sharers(lock_id, _get_lock_num_sharers(lock_id)+increment_by)

def _lock_in_shared_mode(lock_id):
    return _get_lock_num_sharers(lock_id)>0

def _issue_next_lock(lock_id, impose_filesystem_delay=False):
    queue = _get_lock_queue(lock_id)
    if len(queue)>0:
        proc = queue[0][0]
        shared = queue[0][1]
        if shared:
            _issue_shared_locks(lock_id, impose_filesystem_delay)
        else:
            log.logger.debug("Issue lock %r to proc %d", lock_id, proc)
            MessageGrantLock((lock_id, impose_filesystem_delay)).send(proc)

def _issue_shared_locks(lock_id, impose_filesystem_delay=False):
    queue = _get_lock_queue(lock_id)
    sharers_notified = 0
    for proc, shared in queue:
        if shared:
            log.logger.debug("Issue shared lock %r to proc %d",lock_id, proc)
            MessageGrantLock((lock_id, impose_filesystem_delay)).send(proc)
            sharers_notified += 1
    _increment_lock_num_shared(lock_id,sharers_notified)
    log.logger.debug("Lock %r is currently in shared mode, with %d process(es) sharing it",
                     lock_id, _get_lock_num_sharers(lock_id))

def _release_lock_exclusive(lock_id, proc):
    queue = _get_lock_queue(lock_id)
    assert queue[0] == (proc, False)
    queue.pop(0)
    log.logger.debug("Finished with lock %r for proc %d", lock_id, proc)
    if len(queue) > 0:
        _issue_next_lock(lock_id, True)

def _release_lock_shared(lock_id, proc):
    queue = _get_lock_queue(lock_id)
    assert (proc,True) in queue, "Conistency error in locking: can't find a record of the shared lock being released"
    index = queue.index((proc, True))
    del queue[index]
    _increment_lock_num_shared(lock_id, -1)
    log.logger.debug("Finished with shared lock %r for proc %d", lock_id, proc)
    if not _lock_in_shared_mode(lock_id):
        log.logger.debug("Lock %r exiting shared mode", lock_id)
        _issue_next_lock(lock_id)


def _any_locks_alive():
    return any([len(v)>0 for v in _lock_queues.values()])


class ExclusiveLock:
    """Named, exclusive, re-entrant lock - only one MPI process can hold a lock of a given name at once"""
    _shared=False

    def __init__(self, name, delay_before_release=DEFAULT_SLEEP_BEFORE_ALLOWING_NEXT_LOCK):
        self.name = name
        self._delay = delay_before_release
        self._count = 0

    def acquire(self):
        if not parallelism_is_active():
            return
        if self._count==0:
            MessageRequestLock(self.name, self._shared).send(0)
            start = time.time()
            granted = MessageGrantLock.receive(0)
            lock_id, delay = granted.contents
            assert lock_id==self.name, "Received a lock that was not requested. The implementation of ExclusiveLock is not locally thread-safe; are you using multiple threads in one process?"
            if delay:
                time.sleep(self._delay)
            log.logger.debug("Lock %r acquired in %.1fs",self.name, time.time()-start)
        self._count+=1

    def release(self):
        if not parallelism_is_active():
            return
        self._count-=1
        if self._count==0:
            MessageRelinquishLock(self.name).send(0)

    def __enter__(self):
        self.acquire()

    def __exit__(self,type, value, traceback):
        self.release()


class SharedLock(ExclusiveLock):
    """Named, shared, re-entrant lock - multiple MPI processes can hold a lock of a given name at once, but not while an
    ExclusiveLock of the same name is also held"""
    _shared=True



import sys
from typing import Optional

import sqlalchemy
import sqlalchemy.sql.elements
import tqdm
from sqlalchemy import MetaData, create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import Column

from tangos import Base, Creator, DictionaryItem, core
from tangos.config import DB_IMPORT_CHUNK_SIZE
from tangos.core import (HaloLink, HaloProperty, Simulation,
                         SimulationObjectBase, SimulationProperty, TimeStep)

from . import GenericTangosTool


class DBImporter(GenericTangosTool):
    parallel = False
    tool_name = "import"
    tool_description = "Import from one or more different databases (e.g. useful to load sqlite data onto a server)."

    @classmethod
    def add_parser_arguments(self, parser):
        parser.add_argument("files", type=str, nargs="+",
                            help="The filename(s) of the sqlite file, or a sqlalchemy URI, from which to import")
        parser.add_argument("--exclude-properties", type=str, nargs="*", default=[],
                            help="Specify a property that should *excluded* from the copy. "
                                 "Useful if some properties are known to be large.")

    def process_options(self, options):
        self.options = options


    def run_calculation_loop(self):
        self.db_import()

    def db_import(self):
        for remote_db in self.options.files:
            if isinstance(remote_db, sqlalchemy.engine.Engine):
                engine2 = remote_db
            else:
                if "://" not in remote_db:
                    remote_db = "sqlite:///"+remote_db
                engine2 = create_engine(remote_db, echo=False)

            ext_session = sessionmaker(bind=engine2)()

            exclude_dict_ids = [core.get_dict_id(x, session = ext_session) for x in self.options.exclude_properties]
            exclusion_information = {DictionaryItem.__table__.c.id: exclude_dict_ids}
            _db_import_export(core.get_default_session(), ext_session, exclusion_information)


def _db_import_export(target_session, from_session, exclusion_information = None):
    """Copy all database entries from one session into another

    *args*:
    target_session: the session to copy into
    from_session: the session to copy from
    exclusion_information: a dictionary mapping from columns to ids within those columns that should be excluded

    This is a non-trivial operation. The following steps are taken:

    1) For efficiency, all foreign key constraints and indexes are dropped from the target
    2) A copy of the target dictionary is made, which does not have a unique constraint, so that temporary
       duplicates of dictionary entries can be made
    3) All tables are copied from the source to the destination, with the following caveats:
        * The id column of all tables is offset by the existing maximum, to prevent collisions
        * Foreign keys are updated to point to the new ids
        * The dictionary table is copied to the temporary dictionary table, not the permanent one
        * Any sqlalchemy filter expressions in sql_filters are applied. If the table being copied from
          does not have the relevant columns, even after translating primary to foreign keys,
          the filter is ignored.
    4) The temporary dictionary table is de-duplicated, and the result copied back to the permanent dictionary table
    5) Indexes and foreign keys are recreated on the target
    """

    target_connection = target_session.connection()
    from_connection = from_session.connection()

    copy_classes = [Creator, Simulation, TimeStep, SimulationObjectBase, DictionaryItem, SimulationProperty,
                    HaloLink, HaloProperty]

    print("Dropping foreign key constraints...")
    _drop_foreign_keys(target_session)

    print("Dropping indexes...")
    _drop_or_create_indexes(target_connection, mode='drop')

    print("Creating temporary dictionary...")
    # this is necessary because the dictionary table has a unique constraint on the text column
    # which we temporarily need to violate
    temp_dict = _create_temporary_dictionary(target_connection)


    print("Copying tables...")
    try:
        id_offsets = {}
        for target in copy_classes:
            if target == DictionaryItem:
                # special treatment to avoid unique constraint violation - insert into a temporary
                # table and then tidy everything up later
                target_table = temp_dict
            else:
                target_table = None
            id_offsets = _copy_table(from_connection, target_connection, target, id_offsets, target_table,
                                     exclusion_information)

        _dedup_temp_dictionary_items(target_connection, temp_dict)
        _temporary_to_permanent_dictionary(target_connection, temp_dict)

        target_connection.commit()

    finally:
        from_session.close()
        target_connection.rollback()

        print("Recreating indexes...")
        _drop_or_create_indexes(target_connection, mode='create')

        print("Recreating foreign keys...")
        _create_foreign_keys(target_session)

        print("Dropping temporary dictionary table...")
        temp_dict.drop(target_connection)

        target_connection.commit()

        target_session.close()


def _copy_table(from_connection, target_connection, orm_class, offsets, destination_table=None,
                exclusion_information: Optional[dict[Column, list[int]]]=None):
    import tqdm
    from sqlalchemy import func, insert, select

    table = orm_class.__table__

    if destination_table is None:
        destination_table = table

    if exclusion_information is None:
        exclusion_information = {}

    tab = orm_class.__table__

    query_filter = _get_sqlalchemy_filter_from_exclusion_information(exclusion_information, tab)

    num_rows = from_connection.execute(select(func.count(table.c.id)).filter(query_filter)).scalar()

    id_offset = target_connection.execute(select(func.max(table.c.id))).scalar()
    # NB no query_filter should be applied here because we want to know the maximum id in the existing table
    # which may include rows that would be excluded by the filter

    if id_offset is None:
        id_offset = 0

    offsets[table.c.id] = id_offset
    cols_select = _get_import_columns_with_required_offsets(table, offsets)


    num_done = 0

    source_result = from_connection.execute(select(*cols_select).filter(query_filter))

    retries = 0

    with tqdm.tqdm(total=num_rows, desc = f"Copying {orm_class.__name__}", unit="row", smoothing=0.1) as pbar:
        while num_done < num_rows:
            all_rows = source_result.fetchmany(DB_IMPORT_CHUNK_SIZE)
            all_rows = [tuple(r) for r in all_rows]

            try:
                target_connection.execute(insert(destination_table).values(all_rows))
            except sqlalchemy.exc.OperationalError as e:
                if retries>=1:
                    raise # if this line is hit, it may reflect a data limit in the server, e.g. max_allowed_packet in MySQL
                    # Such limits result in the connection being dropped. In PostgreSQL an error is written in the
                    # server log, but in MySQL it does not seem to be. Reducing CHUNK_SIZE may help, or increasing
                    # the limit on the server.

                print(f"Note: lost connection to database after {num_done} rows. Trying again.")
                target_connection.rollback()
                # create a new connection from the target connection's engine
                target_connection = target_connection.engine.connect()
                source_result = from_connection.execute(select(table).offset(num_done))
                retries+=1
                continue

            num_done += len(all_rows)
            pbar.update(len(all_rows))


    return offsets

def _get_foreign_key_dictionary_for_table(table) -> dict[Column, Column]:
    """Return a dictionary mapping foreign columns to local columns for this table"""
    return {fk.column: fk.parent for fk in table.foreign_keys}

def _get_sqlalchemy_filter_from_exclusion_information(exclusion_information: dict[Column, list[int]],
                                                      table: sqlalchemy.schema.Table):
    """Return a sqlalchemy filter that removes rows specified or implied by exclusion_information

     * exclusion_information maps columns to a list of ids that should NOT be included
     * tab is the table for which to generate the filter. Only single-level foreign key relationships
       are mapped; if there are exclusions that are more than one foreign key away, the filter cannot
       be generated (as it would require a join)
    """
    filts = []

    fk_dict = _get_foreign_key_dictionary_for_table(table)
    for column, ids_to_exclude in exclusion_information.items():
        if column.table == table:
            filts.append(~column.in_(ids_to_exclude))
        elif column in fk_dict.keys():
            filts.append(~fk_dict[column].in_(ids_to_exclude))
        else:
            # we can't (currently) handle the case where we are two foreign keys away from the condition,
            # which would necessitate a join. Note that only need to check for being two foreign keys
            # away since a case of "three" foreign jumps necessitates a case elsewhere in the db of
            # two foreign jumps
            for foreign_columns in fk_dict.keys():
                second_level_fks_dict = _get_foreign_key_dictionary_for_table(foreign_columns.table)
                if column in second_level_fks_dict:
                    raise NotImplementedError("This filtering would require a join which is not currently implemented")
    query_filter = sqlalchemy.sql.and_(*filts)
    return query_filter


def _get_import_columns_with_required_offsets(table, offsets):
    cols_select = []
    fk_map = {fk.parent: fk.column for fk in table.foreign_keys}
    for c in table.c:
        if c in fk_map.keys():
            foreign_column = fk_map[c]
        else:
            foreign_column = c

        if foreign_column in offsets.keys() and offsets[foreign_column] != 0:
            cols_select.append(c + offsets[foreign_column])
        else:
            cols_select.append(c)
    return cols_select


def _drop_foreign_keys(session):
    from sqlalchemy import ForeignKeyConstraint, Table
    from sqlalchemy.engine import reflection
    from sqlalchemy.schema import AddConstraint, DropConstraint


    engine = session.get_bind()

    inspector = reflection.Inspector.from_engine(engine)
    fake_metadata = MetaData()

    fake_tables = []
    all_fks = []

    for table_name in Base.metadata.tables:
        # if the metadata has multihop tables left, we aren't interested in those (they're not really in the database)
        if 'multihop' not in table_name:
            fks = []
            for fk in inspector.get_foreign_keys(table_name):
                if fk['name']:
                    fks.append(ForeignKeyConstraint((), (), name=fk['name']))
            t = Table(table_name, fake_metadata, *fks)
            fake_tables.append(t)
            all_fks.extend(fks)

    with engine.begin() as conn:
        for fkc in all_fks:
            print(str(DropConstraint(fkc)))
            conn.execute(DropConstraint(fkc))


def _create_foreign_keys(session, verbose=False):
    from sqlalchemy import ForeignKeyConstraint, MetaData, Table
    from sqlalchemy.engine import reflection
    from sqlalchemy.schema import AddConstraint, DropConstraint


    engine = session.get_bind()

    with engine.begin() as conn:
        for table in Base.metadata.tables.values():
            for fk in table.foreign_keys:
                if verbose:
                    print(str(AddConstraint(fk.constraint)), end="")
                try:
                    conn.execute(AddConstraint(fk.constraint))
                except sqlalchemy.exc.OperationalError:
                    if verbose:
                        print("... FAILED")
                else:
                    if verbose:
                        print("... OK")


def _drop_or_create_indexes(connection, mode='drop', verbose=False):
    from sqlalchemy.schema import CreateIndex, DropIndex
    for table in Base.metadata.tables.values():
        for index in table.indexes:
            try:
                if mode=='drop':
                    if verbose:
                        print(str(DropIndex(index)), end="")
                        sys.stdout.flush()
                    index.drop(connection)
                elif mode=='create':
                    if verbose:
                        print(str(CreateIndex(index)), end="")
                        sys.stdout.flush()
                    index.create(connection)
                else:
                    raise ValueError("mode must be 'drop' or 'create'")
            except (sqlalchemy.exc.OperationalError, sqlalchemy.exc.ProgrammingError):
                if verbose:
                    print("... FAILED")
                connection.rollback()
            else:
                if verbose:
                    print("... OK")
                connection.commit()


def _dedup_temp_dictionary_items(connection, dict_table):
    from sqlalchemy import (Column, bindparam, delete, func, select, text,
                            update)

    updater_expressions = _updates_for_all_columns_referencing_dictionary_id()

    results = connection.execute(select(dict_table.c.text, dict_table.c.id)).fetchall()
    text_to_id = {r[0]: [] for r in results}
    for t, id in results:
        text_to_id[t].append(id)

    text_to_id = {text: id for text, id in text_to_id.items() if len(id)>1 }

    to_delete = []
    for text, id in tqdm.tqdm(text_to_id.items(), desc="De-dup dictionary", unit="items"):
        if len(id)>1:
            update_binds = [{'good_id': id[0], 'bad_id': bad_id} for bad_id in id[1:]]
            for sql in updater_expressions:
                connection.execute(sql, update_binds)
            to_delete+=id[1:]

    connection.execute(delete(dict_table).where(dict_table.c.id.in_(to_delete)))



def _updates_for_all_columns_referencing_dictionary_id():
    from sqlalchemy import bindparam, update

    # find all foreign keys pointing to dictionary.id:
    updates = []
    for table in Base.metadata.tables.values():
        for fk in table.foreign_keys:
            if fk.column == DictionaryItem.__table__.c.id:
                col = fk.parent
                tab = col.table
                update_sql = update(tab).where(col == bindparam('bad_id')).values(
                    **{col.description: bindparam('good_id')})
                updates.append(update_sql)


    return updates




def _temporary_to_permanent_dictionary(connection, temp_dict_table):


    dict_table = DictionaryItem.__table__

    print("Copying back dictionary...")

    # copy the dictionary table into the temporary table
    connection.execute(dict_table.delete())
    connection.execute(dict_table.insert().from_select([dict_table.c.id, dict_table.c.text],
                                                            temp_dict_table.select()))


def _create_temporary_dictionary(connection):
    from sqlalchemy import Column, Integer, String, Table, select
    from sqlalchemy.schema import CreateTable

    dict_table = DictionaryItem.__table__

    # we don't want our table to be associated with the main database schema
    temporary_metadata = MetaData()
    # create a table like the dictionary table, but without the unique constraint
    temp_dict_table = Table("temporary_dictionary", temporary_metadata,
                            Column("id", Integer, primary_key=True),
                            Column("text", String(128)))
    connection.execute(CreateTable(temp_dict_table))

    # copy the dictionary table into the temporary table
    connection.execute(temp_dict_table.insert().from_select([dict_table.c.id, dict_table.c.text],
                                                            dict_table.select()))

    return temp_dict_table



import numpy as np

from .query_masking import QueryMask


class QueryMultivalueFolding:
    """This class manages a situation where a query returns multiple outputs per input, and one temporarily wants
    to explore all those outputs then later fold them back to a single output per input"""

    def __init__(self, determiner_mode, determiner_column, constraints_columns=[]):
        assert determiner_mode in ['max', 'min']
        self.determiner_mode = determiner_mode
        self.determiner_column = determiner_column
        self.constraints_columns = constraints_columns

    def unfold(self, input):
        input_unfolded = []
        slices_for_original_rows = []
        for input_row_index, input_row in enumerate(input):
            startpoint = len(input_unfolded)
            input_unfolded+=input_row
            endpoint = len(input_unfolded)
            slices_for_original_rows.append(slice(startpoint,endpoint))
        self.slices_for_original_rows = slices_for_original_rows
        self.num_original_rows = len(input)
        return input_unfolded

    def refold(self, results_to_refold):
        shape = list(results_to_refold.shape)
        shape[-1] = self.num_original_rows
        out_results = np.empty(shape=shape, dtype=results_to_refold.dtype)
        for i in range(self.num_original_rows):
            results_slice = results_to_refold.T[self.slices_for_original_rows[i]].T
            determiner_slice = results_slice[self.determiner_column]
            #truth_slice = np.full(len(determiner_slice),True).astype(np.bool)
            mask = QueryMask()
            mask.mark_nones_as_masked(determiner_slice)
            results_slice = mask.mask(results_slice.T).T
            determiner_slice = mask.mask(determiner_slice)

            truth_slice = np.full(len(determiner_slice),True).astype(np.bool_)
            for cc in self.constraints_columns:
                constraint = results_slice[cc]
                constraint = mask.mask(constraint)
                constraint = constraint.astype(np.bool_)
                truth_slice[(constraint == False)|(constraint == None)] = False

            if len(determiner_slice)!=0 and True in truth_slice:
                mask.mark_false_as_masked(truth_slice)
                determiner_slice = mask.mask(determiner_slice)
                results_slice = mask.mask(results_slice.T).T
                if self.determiner_mode=='max':
                    select_index = determiner_slice.argmax()
                else:
                    select_index = determiner_slice.argmin()
                out_results_place = out_results.T[i].T
                out_results_place[:] = results_slice.T[select_index].T
        return out_results



from . import YtPropertyCalculation


class StellarMass(YtPropertyCalculation):
    """Returns mass of stars within Rvir of center of halo"""
    names = "Mstar"

    def requires_property(self):
        return ["center", "Rvir_kpc"]

    def calculate(self, particle_data, existing_properties):
        return float(particle_data.quantities.total_quantity(('stars', 'particle_mass')).in_units('Msun').value)



import os
import sys

import tangos.tools

from .. import parallel_tasks
from . import crosslink, manager, timelink, writer


def add_generic_tool(subparse, class_, command, help):
    this_subparser = subparse.add_parser(command, help=help)
    class_.add_parser_arguments(this_subparser)
    def run(options):
        obj = class_()
        obj.process_options(options)
        parallel_tasks.launch(obj.run_calculation_loop, [])
    this_subparser.set_defaults(func=run)

def add_serve_tool(subparse):
    def serve(options):
        from pkg_resources import load_entry_point
        ini_file = options.config
        if os.path.exists(ini_file):
            ini_path = ini_file
        else:
            ini_path = os.path.join(__path__[0],"web",ini_file)
        sys.argv = ["",ini_path]
        sys.exit(
            load_entry_point('pyramid','console_scripts','pserve')()
        )

    web_subparser = subparse.add_parser("serve", help="Start a web server (shortcut to Pyramid's pserve)")
    web_subparser.add_argument('config', action='store', nargs="?",
                               help="The name of the pserve configuration file; either a path or production.ini/development.ini to use tangos' suggested configurations",
                               default="production.ini")
    web_subparser.set_defaults(func=serve)

def add_commands(subparse):
    tangos.tools.GenericTangosTool.add_tools(subparse)
    add_serve_tool(subparse)

def main(argv=None):
    parser, subparse = manager.get_argument_parser_and_subparsers()

    add_commands(subparse)

    args = parser.parse_args(argv)

    from .. import core
    core.process_options(args)
    args.func(args)



from . import PynbodyHaloProperties
from .centring import centred_calculation


class Radius(PynbodyHaloProperties):

    @staticmethod
    def _get_overdensity_contrast():
        raise NotImplementedError("This is meant to be an abstract class")

    @classmethod
    def get_contrast(cls):
        return cls._get_overdensity_contrast()

    @staticmethod
    def _get_reference_definition():
        raise NotImplementedError("This is meant to be an abstract class")

    @classmethod
    def get_rhodef(cls):
        return cls._get_reference_definition()

    @centred_calculation
    def calculate(self, particle_data, existing_properties):
        import pynbody.analysis as analysis
        self._ensure_pynbody_mass_array_loaded_family_level(particle_data)

        # Virial radius is calculated using density contributions form all families
        return analysis.halo.virial_radius(particle_data, overden=self.get_contrast(), rho_def=self.get_rhodef())

    def region_specification(self, existing_properties):
        import pynbody
        return pynbody.filt.Sphere(existing_properties['max_radius'] * 3,
                                   existing_properties['shrink_center'])

    def requires_property(self):
        return ["shrink_center", "max_radius"]

    @staticmethod
    def _ensure_pynbody_mass_array_loaded_family_level(particle_data):
        # Make sure the mass array is loaded at family level
        import pynbody.family
        for family in particle_data.families():
            if family is pynbody.family.dm:
                particle_data.d['mass']
            if family is pynbody.family.star:
                particle_data.s['mass']
            if family is pynbody.family.gas:
                particle_data.g['mass']


class Radius200m(Radius):
    names = "r200m"

    @staticmethod
    def _get_overdensity_contrast():
        return 200

    @staticmethod
    def _get_reference_definition():
        return 'matter'


class Radius200c(Radius):
    names = "r200c"

    @staticmethod
    def _get_overdensity_contrast():
        return 200

    @staticmethod
    def _get_reference_definition():
        return 'critical'


class RadiusVirial(Radius):
    names = "rvirial"

    @staticmethod
    def _get_overdensity_contrast():
        return 178

    @staticmethod
    def _get_reference_definition():
        return 'matter'



from ...input_handlers import pynbody as pynbody_handler_module
from .. import PropertyCalculation


class PynbodyPropertyCalculation(PropertyCalculation):
    works_with_handler = pynbody_handler_module.PynbodyInputHandler
    requires_particle_data = True

PynbodyHaloProperties = PynbodyPropertyCalculation # old name, to be deprecated

from . import BH, SF, centring, eagle, gas, images, mass, profile, radius, zoom



import contextlib
import gc
import sys
import weakref

from ..log import logger


class CheckDeleted:
    def __init__(self, obj):
        if obj is not None:
            self._obj_weakref = weakref.ref(obj)
        else:
            self._obj_weakref = None

    def __enter__(self):
        pass

    def __exit__(self, *_):
        gc.collect()
        if self._obj_weakref is not None and self._obj_weakref() is not None:
            logger.error("check_deleted failed")
            referrers = gc.get_referrers(self._obj_weakref())
            formatted_referrers = ", ".join(["<%s id=%d>" % (type(o), id(o)) for o in referrers])
            logger.error("gc reports hanging references: %s", formatted_referrers)
        self._obj_weakref = None

check_deleted = CheckDeleted



""" Code to read AHF's mtree files """


import os

import numpy as np

from ..log import logger


class AHFTree:
    def __init__(self, path, ts):
        self._path = self._AHF_path_from_snapdir_path(os.path.join(path,ts.extension))
        self._load_fid(ts.previous)
        self._load_raw_links()

    def _AHF_path_from_snapdir_path(cls, path):
        import glob
        ahf_path = path + '*.z*AHF_mtree'
        cat = glob.glob(ahf_path)
        if cat != []:
            return cat[0]
        else:
            raise OSError("Cannot infer path of merger tree files")

    def _load_fid(self, ts):
        self._fid = np.array([x.finder_id for x in ts.halos.all()])

    def _load_raw_links(self):
        """
        There is at least two different file formats for the AHF mtree files. One includes information about the particles shared across haloes,
        the other not and simply orders the progenitors by a merit function.
        This little function is to decide which file format we are dealing with.
        """

        try:
            self._load_mtree_file_standard()
        except:
            logger.info("Could not load AHF mtree file in standard format. Trying the non-standard form.")
            try:
                self._load_mtree_file_cropped()
            except:
                logger.info("Could not load AHF mtree file in non-standard format either. Make sure mtree files exist.")
                raise OSError("Could not load merger tree files")


    def _load_mtree_file_standard(self):
        """
        read in the AHF mtree file containing the indices of halos and its progenitors as well as information about the shared particles.
        We establish symmetric links since for AHF any progenitor can have several descendants cause of mass transfer.
        """
        filename = os.path.join(self._path)
        results = {'id_this':np.asarray([],dtype=np.int64), 'id_desc':np.asarray([],dtype=np.int64), 'f_share':np.asarray([],dtype=np.float64)} #np.empty((0,), dtype=np.dtype([('id_this', np.int64), ('id_desc', np.int64), ('Mvir', np.float32)]))

        data = np.genfromtxt(filename,comments="#",dtype="int")
        i=0
        while i < len(data):
            for j in range(data[i][2]):
                idx = i+1+j
                if data[idx][1] in self._fid: # check if this halo was loaded in case a minimum number of particles different to AHF was used to load halos into DB
                # keep in mind finder id and AHF id have an offset of 1
                    results['id_desc'] = np.append(results['id_desc'],data[i][0])
                    results['id_this'] = np.append(results['id_this'],data[idx][1])
                    results['f_share'] = np.append(results['f_share'], float(data[idx][0] * data[idx][0]) / (data[i][1] * data[idx][2]) )
            i += data[i][2] + 1

        self.links = results

    def _load_mtree_file_cropped(self):
        """
        read in the AHF mtree file containing only the indices of halos and its progenitors and assume progenitors are ordered in descending weight.
        """
        filename = self._path
        results = {'id_this':np.asarray([],dtype=np.int64), 'id_desc':np.asarray([],dtype=np.int64), 'f_share':np.asarray([],dtype=np.float64)} #'Mvir':np.asarray([],dtype=np.float64),

        f = open(filename)
        lines = f.readlines()
        nhalos = int(lines[0])
        skip = 1 #skip lines already read
        i=0
        while i < nhalos:
            i += 1
            _tmp = np.fromstring(lines[skip], dtype=np.dtype(int,int), sep=' ')
            _id = int(_tmp[0])
            nprogen = int(_tmp[1])
            skip += 1 # increment the skip of lines for the line read above
            if nprogen > 0:
                for n in range(nprogen):
                    _this_id = int(lines[skip+n]) # rip off the timestep which is encoded as the first 3 digits
                    if _this_id in self._fid: # check if the halo exists in the database, this is needed if db was created with a minimum particle number per halo which does not agree with AHF definition
                        results['id_desc'] = np.append(results['id_desc'],np.asarray([_id],dtype=np.int64))
                        results['id_this'] = np.append(results['id_this'],np.asarray([_this_id],dtype=np.int64))
                        results['f_share'] = np.append(results['f_share'], np.asarray([(nprogen-n)/nprogen],dtype=np.float64))
            skip += nprogen   # increment line skip by already read lines
        self.links = results

    def _load_major_progenitor_branch(self):
        """
        load the major progenitor branch of the merger tree from the mtree_idx files.
        """
        NotImplementedError("Loading the major progenitor branch is not implemented")


    def get_links_for_snapshot(self):
        """Get the links from snapshot ts to its immediate successor.

        Returns a dictionary; keys are the finder IDs at the given snapnum, whereas the values are
        a tuple containing the ID at the subsequent snapshot and the fraction of shared particles.
        """
        ids_this_snap = self.links['id_this']
        ids_next_snap = self.links['id_desc']
        merger_ratios = self.links['f_share']

        return list(zip(ids_this_snap, zip(ids_next_snap, merger_ratios)))



import sqlalchemy.engine

from ..log import logger


def explain_query(query, engine_or_connection=None):
    """Get the underlying SQL engine to explain how it will execute a given query. For debugging purposes.

    If engine_or_connection is None, use the query's existing engine"""

    if engine_or_connection is None:
        engine_or_connection = query.session.connection()

    if isinstance(engine_or_connection, sqlalchemy.engine.Engine):
        with engine_or_connection.connect() as connection:
            _explain_query_using_connection(query, connection)
    else:
        _explain_query_using_connection(query, engine_or_connection)

def _explain_query_using_connection(query, connection):
    compiled_q = query.statement.compile(compile_kwargs={"literal_binds": True})
    # use self._connection.execute("explain "+str(compiled_q)) to get explanation, then print it:

    from sqlalchemy.sql import text
    dialect = connection.dialect.dialect_description.split("+")[0].lower()
    if dialect == 'sqlite':
        explain_command = "explain query plan "
    else:
        explain_command = "explain analyze "

    explain_result = connection.execute(text(explain_command + str(compiled_q)))

    from prettytable import from_db_cursor


    logger.info("Analysis of query:")
    logger.info(compiled_q)

    pt = from_db_cursor(explain_result.cursor)
    pt.align = "l"


    logger.info(pt)



import numbers

import numpy as np

from .. import config, core
from .. import parallel_tasks as pt
from ..core import Simulation, TimeStep
from ..log import logger


class SimulationAdderUpdater:
    """This class contains the necessary tools to add a new simulation to the database"""

    def __init__(self, simulation_output, renumber=True):
        """:type simulation_output tangos.simulation_outputs.HandlerBase"""
        self.simulation_output = simulation_output

        # if running in parallel, creating a session for the first time may trigger a race condition e.g. if the
        # database doesn't exist yet
        with pt.ExclusiveLock("creating_database"):
            core.get_default_session()

        self.min_halo_particles = config.min_halo_particles
        self.max_num_objects = config.max_num_objects
        self.renumber = renumber

        self.parallel = pt.parallelism_is_active()


    @property
    def session(self):
        return core.get_default_session()

    @property
    def basename(self):
        return self.simulation_output.basename

    def scan_simulation_and_add_all_descendants(self):
        if self.parallel:
            assert pt.parallelism_is_active(), "Parallel backend has not been initialized"
            from ..parallel_tasks import database
            database.synchronize_creator_object()
            create_simulation = pt.backend.rank()==1 # nb rank 0 is busy coordinating everything
        else:
            create_simulation = True

        if create_simulation:
            if not self.simulation_exists():
                logger.info("Add new simulation %r", self.basename)
                logger.info("... using the output handler %r", self.simulation_output.handler_class_name())
                self.add_simulation()
            else:
                logger.warning("Simulation already exists %r", self.basename)

            self.add_simulation_properties()

        # await the simulation being ready if we are running in parallel
        pt.barrier()

        for ts_filename in self.simulation_output.enumerate_timestep_extensions(parallel=self.parallel):
            if not self.timestep_exists_for_extension(ts_filename):
                ts = self.add_timestep(ts_filename)
                self.add_timestep_properties(ts)
                self.add_objects_to_timestep(ts, core.halo.Halo)
                self.add_objects_to_timestep(ts, core.halo.Group)
            else:
                logger.warning("Timestep already exists %r", ts_filename)

    def simulation_exists(self):
        num_matches = self.session.query(Simulation).filter_by(basename=self.basename).count()
        assert num_matches<2, "Consistency problem - more than one simulation with this name exists"
        return num_matches>0

    def timestep_exists_for_extension(self, ts_extension):
        with pt.lock.SharedLock("db_write_lock"):
            ex = core.get_default_session().query(TimeStep).filter_by(
                simulation=self._get_simulation(),
                extension=ts_extension).first()
        return ex is not None

    def add_timestep(self, ts_extension):
        logger.info("Add timestep %r to simulation %r",ts_extension,self.basename)
        ex = TimeStep(self._get_simulation(), ts_extension)
        with pt.ExclusiveLock("db_write_lock"):
            self.session.add(ex)
            self.session.commit()
        return ex

    def add_simulation(self):
        sim = Simulation(self.basename)
        # no need for a lock here - only one process should be adding a simulation
        self.session.add(sim)
        self.session.commit()

    def add_simulation_properties(self):
        sim = self._get_simulation()
        properties_dict = self.simulation_output.get_properties()
        properties_dict['handler'] = self.simulation_output.handler_class_name()
        for k, v in properties_dict.items():
            if isinstance(v, numbers.Number) and np.isnan(v):
                continue
            if k not in list(sim.keys()):
                logger.info("Add simulation property %r",k)
                sim[k] = v
            elif sim[k]!=v:
                logger.info("Update simulation property %r", k)
                sim[k] = v
            else:
                logger.warning("Simulation property %r already exists", k)

        self.session.commit()

    @staticmethod
    def _autoadd_zeros(enumerate_fn):
        # enable enumerate functions to just return halo numbers, adding on the number of dark matter, star and
        # gas particles for backwards compatibility
        def adapted(*args, **kwargs):
            for result in enumerate_fn(*args, **kwargs):
                if not hasattr(result, "__len__"):
                    yield result, result, 0, 0, 0
                else:
                    yield result
        return adapted

    def add_objects_to_timestep(self, ts, create_class=core.halo.Halo):
        halos = []
        n_tot = []
        enumerator = self._autoadd_zeros(self.simulation_output.enumerate_objects)

        for catalog_id, finder_id, NDM, Nstar, Ngas in enumerator(ts.extension, object_typetag=create_class.tag,
                                                      min_halo_particles=self.min_halo_particles):
            n_tot.append(NDM+Nstar+Ngas)

        if self.renumber:
            database_id = np.zeros(len(n_tot), dtype=int)

            # Sort by total particle number, largest objects first. Use mergesort for sort stability.
            database_id[np.argsort(-np.array(n_tot),kind='mergesort')] = np.arange(len(n_tot)) + 1
        else:
            database_id = [None]*len(n_tot)

        for database_number, (catalog_id, finder_id, NDM, Nstar, Ngas) in zip(database_id,
                                                                 enumerator(ts.extension, object_typetag=create_class.tag,
                                                                            min_halo_particles=self.min_halo_particles)):
            if database_number is None:
                database_number = catalog_id

            if (NDM+Nstar+Ngas >= self.min_halo_particles or NDM==0) \
                    and (self.max_num_objects is None or database_number<=self.max_num_objects ):
                h = create_class(ts, database_number, finder_id, catalog_id, NDM, Nstar, Ngas)
                halos.append(h)

        with pt.ExclusiveLock("db_write_lock"):
            logger.info("Add %d %ss to timestep %r", len(halos), create_class.__name__, ts)
            self.session.add_all(halos)
            self.session.commit()

    def add_timestep_properties(self, ts):
        for key, value in self.simulation_output.get_timestep_properties(ts.extension).items():
            setattr(ts, key, value)



    def _get_simulation(self):
        with pt.lock.SharedLock("db_write_lock"):
            return self.session.query(Simulation).filter_by(basename=self.basename).first()



import logging
import time

from ..config import PROPERTY_WRITER_PARALLEL_STATISTICS_TIME_BETWEEN_UPDATES
from ..log import logger
from .message import Message

_new_accumulator_requested_for_ranks = []
_new_accumulator = None
_existing_accumulators = []
class CreateNewAccumulatorMessage(Message):

    def process(self):
        from . import backend
        global _new_accumulator, _new_accumulator_requested_for_ranks, _existing_accumulators
        assert issubclass(self.contents, StatisticsAccumulatorBase)
        if _new_accumulator is None:
            _new_accumulator = self.contents()
            _new_accumulator_requested_for_ranks = [self.source]
        else:
            assert self.source not in _new_accumulator_requested_for_ranks
            assert isinstance(_new_accumulator, self.contents)
            _new_accumulator_requested_for_ranks.append(self.source)

            from . import backend

        if len(_new_accumulator_requested_for_ranks) == backend.size()-1:
            self._confirm_new_accumulator()

    def _confirm_new_accumulator(self):
        global _new_accumulator, _new_accumulator_requested_for_ranks, _existing_accumulators
        from . import backend, on_exit_parallelism
        accumulator_id = len(_existing_accumulators)
        _existing_accumulators.append(_new_accumulator)

        locally_bound_accumulator = _new_accumulator
        logger.debug("Created new accumulator of type %s with id %d" % (locally_bound_accumulator.__class__.__name__, accumulator_id))
        on_exit_parallelism(lambda: locally_bound_accumulator.report_to_log_if_needed(logger, 0.05))

        _new_accumulator = None
        _new_accumulator_requested_for_ranks = []

        for destination in range(1, backend.size()):
            AccumulatorIdMessage(accumulator_id).send(destination)
class AccumulatorIdMessage(Message):
    pass
class AccumulateStatisticsMessage(Message):
    def process(self):
        global _existing_accumulators
        _existing_accumulators[self.contents.id].add(self.contents)

class StatisticsAccumulatorBase:
    REPORT_AFTER = PROPERTY_WRITER_PARALLEL_STATISTICS_TIME_BETWEEN_UPDATES
    def __init__(self, allow_parallel=False):
        """This is a base class for accumulating statistics, possibly in parallel across multiple processes.

        Note that if allow_parallel is True, then all processes must create an instance of this class
        (in effect creating the class will act like a barrier). If only some processes create an instance,
        they will block, possibly creating a deadlock. This is why allow_parallel defaults to False.
        """
        from . import backend, parallelism_is_active
        self._last_reported = time.time()
        self._parallel = allow_parallel and parallelism_is_active() and backend.rank() != 0
        if self._parallel:
            logger.debug(f"Registering {self.__class__}")
            CreateNewAccumulatorMessage(self.__class__).send(0)
            logger.debug(f"Awaiting accumulator id for {self.__class__}")
            self.id = AccumulatorIdMessage.receive(0).contents
            logger.debug(f"Received accumulator id={ self.id}")

    def report_to_server(self):
        if self._parallel:
            AccumulateStatisticsMessage(self).send(0)
            self.reset()

    def reset(self):
        raise NotImplementedError("This method should be overriden to reset the statistics")

    def add(self, other):
        raise NotImplementedError("This method should be overriden to add two accumulations together")

    def report_to_log(self, logger):
        raise NotImplementedError("This method should be overriden to log a statistics report")

    def report_to_log_or_server(self, logger):
        if self._parallel:
            self.report_to_server()
        else:
            self.report_to_log(logger)

    def report_to_log_if_needed(self, logger, after_time=None):
        if after_time is None:
            after_time = self.REPORT_AFTER
        if time.time() - self._last_reported > after_time:
            self.report_to_log(logger)
            self._last_reported = time.time()
